/* SPDX-FileCopyrightText: 2018-2020 Tobias Leupold <tobias.leupold@gmx.de>

   SPDX-License-Identifier: GPL-3.0-or-later
*/

// Local includes

#include "Client.h"

#include "RegistrationPage/RegistrationPage.h"
#include "RegistrationPage/MarkersWidget.h"

#include "shared/Tournament.h"
#include "shared/SharedObjects.h"
#include "shared/Settings.h"
#include "shared/TournamentSettings.h"
#include "shared/StringTools.h"
#include "shared/JsonHelper.h"

// Qt includes
#include <QDebug>
#include <QHostAddress>
#include <QTcpSocket>
#include <QTimer>
#include <QJsonObject>
#include <QJsonArray>
#include <QColor>
#include <QApplication>
#include <QLocale>

namespace sp = ServerProtocol;

Client::Client(QObject *parent, int connectTimeout, int requestTimeout,
               SharedObjects *sharedObjects, RegistrationPage *registrationPage)
    : AbstractNetworkEngine(parent),
      m_connectTimeout(connectTimeout),
      m_db(sharedObjects->database()),
      m_settings(sharedObjects->settings()),
      m_tournamentSettings(sharedObjects->tournamentSettings()),
      m_stringTools(sharedObjects->stringTools()),
      m_registrationPage(registrationPage),
      m_markersWidget(registrationPage->markersWidget()),
      m_requestHandlers {
          // Registration changes

          { sp::registerPlayers,         &Client::processRegisterPlayers },
          { sp::renamePlayers,           &Client::processRenamePlayers },
          { sp::assemblePair,            &Client::processAssemblePair },
          { sp::deletePlayers,           &Client::processDeletePlayers },
          { sp::updatePlayersMarker,     &Client::processUpdatePlayersMarker },

          { sp::deleteMarkedPlayers,     &Client::processDeleteMarkedPlayers },
          { sp::deleteUnmarkedPlayers,   &Client::processDeleteUnmarkedPlayers },
          { sp::deleteAllPlayers,        &Client::processDeleteAllPlayers },
          { sp::updatePlayersListMarker, &Client::processUpdateListMarker },
          { sp::removePlayersListMarker, &Client::processRemoveListMarker },

          // Draw changes
          { sp::updateDraw,     &Client::processUpdateDraw },
          { sp::setListDraw,    &Client::processSetListDraw },
          { sp::deleteAllDraws, &Client::processDeleteAllDraws },

          // Markers definition changes
          { sp::addMarker,    &Client::processAddMarker },
          { sp::deleteMarker, &Client::processDeleteMarker },
          { sp::moveMarker,   &Client::processMoveMarker },
          { sp::editMarker,   &Client::processEditMarker },

          // Special markers changes
          { sp::setAbsentMarker,         &Client::processSetAbsentMarker },
          { sp::singlesManagementChange, &Client::processSinglesManagementChange },

          // Finishing the registration
          { sp::finishRegistration, &Client::processFinishRegistration },

          // Ranking comparison
          { sp::rankingRequest, &Client::processRankingRequest },

          // Score change
          { sp::scoreChange, &Client::processScoreChange },

          // Score comparison
          { sp::scoreRequest, &Client::processScoreRequest }
    }
{
    connect(socket(), QOverload<QAbstractSocket::SocketError>::of(&QAbstractSocket::error),
            this, &Client::error);
    connect(socket(), &QAbstractSocket::connected, this, &Client::socketConnected);

    m_connectionTimer = new QTimer(this);
    m_connectionTimer->setInterval(connectTimeout);
    m_connectionTimer->setSingleShot(true);
    connect(m_connectionTimer, &QTimer::timeout, this, &Client::connectionTimeout);

    m_requestTimer = new QTimer(this);
    m_requestTimer->setInterval(requestTimeout);
    m_requestTimer->setSingleShot(true);
    connect(m_requestTimer, &QTimer::timeout, this, &Client::requestTimeout);
}

void Client::setChecksum(const QString *checksum)
{
    m_checksum = checksum;
}

void Client::connectToServer(const QHostAddress &address, quint16 port)
{
    emit message(tr("Verbinde mit Server %1").arg(address.toString()));
    m_connectionPending = true;
    m_connectionTimer->start();
    emit message(tr("Stelle Socketverbindung her …"));
    socket()->connectToHost(address, port);
}

QString Client::serverIp() const
{
    return socket()->peerAddress().toString();
}

int Client::serverPort() const
{
    return socket()->peerPort();
}

void Client::socketConnected()
{
    emit message(tr("Socketverbindung hergestellt, initialisiere Handshake …"));
    m_connectionPending = false;
    m_handshakePending = true;
}

void Client::abortHandshake()
{
    if (! m_handshakePending) {
        return;
    }

    closeConnection();
    m_connectionTimer->stop();
    m_handshakePending = false;
}

void Client::error(QAbstractSocket::SocketError)
{
    closeConnection();

    if (m_connectionPending || m_handshakePending) {
        m_connectionTimer->stop();
        m_handshakePending = false;
        m_connectionPending = false;
        emit connectResult(false, socket()->errorString());
        return;
    }

    emit connectionError(socket()->errorString());
}

void Client::connectionTimeout()
{
    if (m_connectionPending) {
        emit connectionError(tr("Konnte keine Socketverbindung herstellen."));
        return;
    }

    closeConnection();
    QLocale locale;
    emit message(tr("Timeout: Keine Antwort nach %1 Sekunden").arg(
                 locale.toString(float(m_connectTimeout) / 1000.0)));
    emit connectResult(false, tr("Der Server hat zu lange keine Antwort geschickt."));
}

void Client::closeConnection()
{
    socket()->abort();
    m_connected = false;
}

void Client::sendRequest(QJsonObject &jsonObject, ProcessingTarget target,
                         const QString &request, const QString &scrollToTarget)
{
    emit tx();
    emit message(tr("Frage %1 an").arg(request));
    emit statusUpdate(tr("Frage %1 beim Server an …").arg(request));

    switch (target) {
    case ProcessingTarget::None:
        break;
    case ProcessingTarget::Registration:
        m_registrationPage->setProcessingRequest(true, scrollToTarget);
        break;
    case ProcessingTarget::Markers:
        m_markersWidget->setProcessingRequest(true);
        break;
    }

    jsonObject.insert(sp::checksum, *m_checksum);
    jsonObject.insert(sp::clientId, m_id);
    sendData(jsonObject);

    m_processingTarget.append(target);
    m_requestTimer->start();
}

int Client::id() const
{
    return m_id;
}

void Client::requestTimeout()
{
    m_registrationPage->setProcessingRequest(false);
    m_markersWidget->setProcessingRequest(false);
    emit connectionError(tr("Der Server hat zu lange nicht auf eine Anfrage reagiert. Bitte die "
                            "Netzwerkverbindung überprüfen und neu verbinden."));
    closeConnection();
}

bool Client::connected() const
{
    return m_connected;
}

void Client::stopProcessingRequest()
{
    m_requestTimer->stop();
    switch (m_processingTarget.takeFirst()) {
    case ProcessingTarget::None:
        break;
    case ProcessingTarget::Registration:
        m_registrationPage->setProcessingRequest(false);
        break;
    case ProcessingTarget::Markers:
        m_markersWidget->setProcessingRequest(false);
        break;
    }
}

void Client::verifyChecksum(const QString &checksum)
{
    if (checksum != *m_checksum) {
        // This should not happen. But who knows?!
        if (m_processingTarget.count() > 0) {
            stopProcessingRequest();
        }
        emit connectionError(tr("Die Synchronisation mit dem Server ist fehlgeschlagen, die Daten "
                                "weichen ab. Die Verbindung wird beendet. Bitte neu verbinden, um "
                                "den Datenbestand wieder mit dem Server abzugleichen!"));
    }
}

// ====================
// Handshake processing
// ====================

void Client::processHandshake(const QJsonObject &jsonObject)
{
    const QJsonValue phaseValue = jsonObject.value(sp::phase);
    if (phaseValue.type() != QJsonValue::String) {
        emit connectResult(false, tr("Der Server hat eine ungültige Begrüßung geschickt"));
        return;
    }

    const QString handshakePhase = phaseValue.toString();

    // Handshake Initialization
    // ========================

    if (handshakePhase == sp::initialization) {

        const QJsonValue serverSignatureValue = jsonObject.value(sp::serverSignatureKey);
        if (serverSignatureValue.type() != QJsonValue::String
            || serverSignatureValue.toString() != sp::serverSignature) {

            emit message(tr("Der Server hat eine ungültige Begrüßung geschickt."));
            emit connectResult(false, tr("Der Server hat eine ungültige Begrüßung geschickt."));
            return;
        }

        const QJsonValue protocolVersionValue = jsonObject.value(sp::protocolVersionKey);
        if (protocolVersionValue.type() != QJsonValue::Double) {
            emit message(tr("Der Server hat eine ungültige Begrüßung geschickt."));
            emit connectResult(false, tr("Der Server hat eine ungültige Begrüßung geschickt."));
            return;
        } else {
            const int protocolVersion = protocolVersionValue.toInt();
            if (protocolVersion != sp::protocolVersion) {
                emit message(tr("Falsche Protokollrevision (ist %1, müsste %2 sein)").arg(
                                QString::number(protocolVersion),
                                QString::number(sp::protocolVersion)));
                emit connectResult(false, tr(
                    "Der Server verwendet eine andere Protokollrevision (ist %1, müsste %2 sein). "
                    "Es kann keine Verbindung hergestellt werden.").arg(
                    QString::number(protocolVersion), QString::number(sp::protocolVersion)));
                return;
            }
        }

        // The counterpart seems to be a Muckturnier server and we speak the same protocol revision
        emit message(tr("Begrüßung empfangen, frage Synchronisation an …"));
        sendData(sp::synchronizationRequest);

    // Data Synchronization
    // ===================

    } else if (handshakePhase == sp::synchronization) {

        // Check the settings and state
        // ----------------------------

        const QJsonObject settings          = jsonObject.value(sp::settings).toObject();
        const int boogerScore               = settings.value(sp::boogerScore).toInt();
        const int boogersPerRound           = settings.value(sp::boogersPerRound).toInt();
        const bool networkTournamentStarted = settings.value(sp::tournamentStarted).toBool();
        const Tournament::Mode tournamentMode
            = static_cast<Tournament::Mode>(settings.value(sp::tournamentMode).toInt());

        // Tournament settings
        if (m_tournamentSettings->tournamentMode() != tournamentMode
            || m_tournamentSettings->boogerScore() != boogerScore
            || m_tournamentSettings->boogersPerRound() != boogersPerRound) {

            emit message(tr("Die Datenbank des Servers hat andere Einstellungen als die lokale"));
            emit connectResult(false, tr(
                "Die Datenbank des Servers hat andere Einstellungen als die lokale:</p>"
                "<table>"
                "<tr>"
                    "<td></td>"
                    "<td>Server</td>"
                    "<td>Hier</td>"
                "</tr>"
                "<tr>"
                    "<td>Turniermodus:</td>"
                    "<td><i>%1</i>&nbsp;&nbsp;</td>"
                    "<td><i>%2</i></td>"
                "</tr>"
                "<tr>"
                    "<td>Punkte pro Bobbl:&nbsp;&nbsp;</td>"
                    "<td><i>%3</i></td>"
                    "<td><i>%4</i></td>"
                "</tr>"
                "<tr>"
                    "<td>Bobbl pro Runde:</td>"
                    "<td><i>%5</i></td>"
                    "<td><i>%6</i></td>"
                "</tr>"
                "</table>"
                "<p>Es kann keine Verbindung hergestellt werden.").arg(
                tournamentMode == Tournament::Mode::FixedPairs ? tr("Feste Paare")
                                                               : tr("Einzelne Spieler"),
                m_tournamentSettings->isPairMode() ? tr("Feste Paare") : tr("Einzelne Spieler"),
                QString::number(boogerScore),
                QString::number(m_tournamentSettings->boogerScore()),
                QString::number(boogersPerRound),
                QString::number(m_tournamentSettings->boogersPerRound())));
            return;
        }

        // Registration phase state
        if (m_db->localTournamentStarted() && ! networkTournamentStarted) {
            emit message(tr("Es wurden bereits Ergebnisse eingegeben, aber nicht auf dem Server"));
            emit connectResult(false,
                     tr("Es wurden bereits Ergebnisse eingegeben, aber der Server hat die "
                        "Anmeldung noch nicht beendet.</p>"
                        "<p>Bitte nach Beendigung der Anmeldung erneut verbinden, oder die bisher "
                        "eingegebenen Ergebnisse löschen und erneut verbinden!"));
            return;
        }

        // Process the Server's data
        // -------------------------

        // Extract the markers
        QVector<Markers::Marker> markers;
        const auto markersArray = jsonObject.value(sp::markersList).toArray();
        for (const QJsonValue &value : markersArray) {
            const QJsonObject entry = value.toObject();
            Markers::Marker marker;
            marker.id    = entry.value(sp::markerId).toInt();
            marker.name  = entry.value(sp::markerName).toString();
            marker.color = colorFromValue(entry.value(sp::markerColor));
            marker.sorting
                = static_cast<Markers::MarkerSorting>(entry.value(sp::markerSorting).toInt());
            markers.append(marker);
        }

        // Extract the players data
        QHash<QString, Database::PlayersData> playersData;
        const auto jsonValuePlayers = jsonObject.value(sp::playersList).toArray();
        for (const QJsonValue &value : jsonValuePlayers) {
            const QJsonObject entry    = value.toObject();
            const QString name         = entry.keys().at(0); // This is save, keys() is never empty
            const QJsonObject jsonData = entry.value(name).toObject();
            Database::PlayersData data;
            data.marker             = jsonData.value(sp::markerId).toInt();
            data.assignedTable      = jsonData.value(sp::playersTableNumber).toInt();
            data.assignedPairNumber = jsonData.value(sp::playersPairNumber).toInt();
            playersData[name] = data;
        }
        QStringList players(playersData.keys());

        // Adopt the data
        // --------------

        // Adopt the locale
        const QString serverLocale = settings.value(sp::locale).toString();
        if (m_stringTools->locale() != serverLocale) {
            emit message(tr("Übernehme abweichendes Server-Gebietsschema „%1“ (die lokale "
                            "Einstellung ist „%2“)").arg(serverLocale, m_stringTools->locale()));
            m_stringTools->setLocale(serverLocale);
        }

        // Adopt the registration phase state
        m_db->setNetworkTournamentStarted(networkTournamentStarted);

        // If we don't have results yet, we can simply adopt all data
        if (! m_db->localTournamentStarted()) {
            if (! m_db->adoptAllData(playersData, markers)) {
                return;
            }
            m_registrationPage->reload();

        // Otherwise, the local players list has to be matched and probably updated
        } else {
            auto [ localPlayers, success ] = m_db->players();
            if (! success) {
                return;
            }

            if (localPlayers.count() != players.count()) {
                emit message(tr("Es wurden bereits Ergebnisse eingetragen, und es sind "
                                "unterschiedlich viele %1 angemeldet").arg(
                                m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));
                emit connectResult(false, tr(
                    "Es wurden bereits Ergebnisse eingetragen, und es sind unterschiedlich viele "
                    "%1 angemeldet (am Server %2, hier %3). Es kann keine Verbindung hergestellt "
                    "werden!").arg(m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                                   QString::number(players.count()),
                                   QString::number(localPlayers.count())));
                return;
            }

            m_stringTools->sort(localPlayers);
            m_stringTools->sort(players);
            if (localPlayers == players) {
                // The lists match, we can leave it as-is and adopt the markers
                if (! m_db->adoptMarkersAndDraw(playersData, markers)) {
                    return;
                }
                m_registrationPage->reload();
            } else {
                // The names differ, so we have to match the lists first
                emit namesDiffer(localPlayers, players);
                emit connectResult(false, tr(
                    "Es wurden bereits Ergebnisse eingetragen, und die %1 des Servers "
                    "unterscheidet sich von der lokalen. Bevor eine Verbindung hergestellt werden "
                    "kann, müssen alle Namen übereinstimmen.</p>"
                    "<p>Bitte die Liste überprüfen, die Namen übernehmen bzw. manuell korrigieren "
                    "und erneut verbinden!").arg(
                    m_tournamentSettings->isPairMode() ? tr("Paarliste") : tr("Spielerliste")));
                return;
            }
        }

        // Adopt the absent marker
        const int absentMarkerId = jsonObject.value(sp::absentMarker).toInt();
        m_markersWidget->setCheckableMarker(Markers::Absent, absentMarkerId);
        m_tournamentSettings->setMarker(Markers::Absent, absentMarkerId);

        if (m_tournamentSettings->isPairMode()) {
            // Adopt the singles management settings
            const Markers::SinglesManagement singlesManagement
                = JsonHelper::parseSinglesManagement(jsonObject.value(
                      sp::singlesManagement).toObject());
            m_markersWidget->setSinglesManagement(singlesManagement);
            m_tournamentSettings->setSinglesManagement(singlesManagement);
        }

        if (m_tournamentSettings->isSinglePlayerMode()) {
            // Adopt the server's names separator
            m_settings->setTransientNamesSeparator(settings.value(sp::namesSeparator).toString());
        }

        // Update the players list to show the adopted settings
        m_registrationPage->updatePlayersList();

        // We have to invoke RegistrationPage::updateChecksum() here manually, because as of now,
        // the registration page does not know yet that this instance is a client and won't cache
        // the checksum whilst running updatePlayersList()
        m_registrationPage->updateChecksum();

        // Set the initial remote score checksums
        const QJsonObject scoreChecksums
            = jsonObject.value(sp::scoreChecksumsList).toObject();
        const auto rounds = scoreChecksums.keys();
        for (const auto &round : rounds) {
            emit remoteChecksumChanged(round.toInt(), scoreChecksums.value(round).toString());
        }

        // Here, the data should be in sync
        if (jsonObject.value(sp::checksum).toString() == *m_checksum) {
            // We are in sync with the server and everything seems to be fine, so we want an ID
            emit message(tr("Synchronisation war erfolgreich, frage ID an …"));
            sendData(sp::idAllocationRequest);
        } else {
            // This should not happen, but who knows?!
            emit message(tr("Die Übernahme der Daten des Servers ist fehlgeschlagen!"));
            emit connectResult(false,
                               tr("Die Übernahme der Daten des Servers ist fehlgeschlagen!"));
            return;
        }

    // Finish the handshake
    // ====================

    } else if (handshakePhase == sp::idAllocation) {
        m_connectionTimer->stop();
        m_handshakePending = false;
        m_id = jsonObject.value(sp::clientId).toInt();
        emit message(tr("Verbindung mit %1 Port %2 hergestellt, Client-ID: %3").arg(
                        serverIp(), QString::number(serverPort()), QString::number(m_id)));
        emit connectResult(true);
        m_connected = true;
    }
}

void Client::statusMessage(const QString &message, int clientId)
{
    if (clientId == m_id) {
        emit statusUpdate(message);
    } else {
        emit statusUpdate(tr("Netzwerkänderung: ") + message);
    }
}

// ======================
// Requests to the server
// ======================

// Registration changes
// ====================

void Client::requestRegisterPlayers(const QString &name, int marker)
{
    QJsonObject message { { sp::action,      sp::registerPlayers },
                          { sp::playersName, name },
                          { sp::markerId,    marker } };
    sendRequest(message, ProcessingTarget::Registration,
                tr("Anmelden von „%1“").arg(name),
                name);
}

void Client::requestRenamePlayers(const QString &name, const QString &newName,
                                  int markerId, const QString &markerName)
{
    QJsonObject message { { sp::action,         sp::renamePlayers },
                          { sp::playersName,    name },
                          { sp::newPlayersName, newName },
                          { sp::markerId,       markerId },
                          { sp::markerName,     markerName } };

    QString markerChange;
    if (markerId != -1) {
        if (markerId == 0) {
            markerChange = tr(" und Entfernen der Markierung");
        } else {
            markerChange = tr(" und Setzen der Markierung auf „%1“").arg(markerName);
        }
    }
    sendRequest(message, ProcessingTarget::Registration,
                tr("Umbenennen von „%1“ zu „%2“%3").arg(name, newName, markerChange),
                newName);
}

void Client::requestAssemblePair(const QString &player1, const QString &player2,
                                 const QString &separator, int assignedMarker)
{
    QJsonObject message { { sp::action,      sp::assemblePair },
                          { sp::playersName, QJsonArray { player1, player2, separator } },
                          { sp::markerId,    assignedMarker } };
    sendRequest(message, ProcessingTarget::Registration,
                tr("Zusammenfügen von „%1“ und „%2“ zu einem Paar").arg(player1, player2),
                player1 + separator + player2);
}

void Client::requestDeletePlayers(const QString &name)
{
    QJsonObject message { { sp::action,      sp::deletePlayers },
                          { sp::playersName, name } };
    sendRequest(message, ProcessingTarget::Registration,
                tr("Löschen von „%1“").arg(name));
}

void Client::requestUpdateMarker(const QString &name, const QString &markerName, int markerId)
{
    QJsonObject message { { sp::action,      sp::updatePlayersMarker },
                          { sp::playersName, name },
                          { sp::markerName,  markerName },
                          { sp::markerId,    markerId } };
    if (markerId != 0) {
        sendRequest(message, ProcessingTarget::Registration,
                    tr("Markieren von „%1“ als „%2“ an").arg(name, markerName),
                    name);
    } else {
        sendRequest(message, ProcessingTarget::Registration,
                    tr("Entfernen der Markierung von „%1“ an").arg(name),
                    name);
    }
}

void Client::requestDeleteMarkedPlayers(const QString &markerName, int markerId)
{
    QJsonObject message { { sp::action,     sp::deleteMarkedPlayers },
                          { sp::markerName, markerName },
                          { sp::markerId,   markerId } };
    sendRequest(message, ProcessingTarget::Registration,
                tr("Löschen aller mit „%1“ markierten %2").arg(
                   markerName, m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));
}

void Client::requestDeleteUnmarkedPlayers()
{
    QJsonObject message { { sp::action, sp::deleteUnmarkedPlayers } };
    sendRequest(message, ProcessingTarget::Registration,
                tr("Frage Löschen aller unmarkierten %1").arg(
                   m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));
}

void Client::requestDeleteAllPlayers()
{
    QJsonObject message { { sp::action, sp::deleteAllPlayers } };
    sendRequest(message, ProcessingTarget::Registration,
                tr("Löschen aller %1 an").arg(
                   m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));
}

void Client::requestListUpdateMarker(const QString &markerName, int markerId)
{
    QJsonObject message { { sp::action,     sp::updatePlayersListMarker },
                          { sp::markerName, markerName },
                          { sp::markerId,   markerId } };

    if (markerId != 0) {
        sendRequest(message, ProcessingTarget::Registration,
                    tr("Markieren aller %1 als „%2“ an").arg(
                       m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                       markerName));
    } else {
        sendRequest(message, ProcessingTarget::Registration, tr("Entfernen aller Markierungen"));
    }
}

void Client::requestListRemoveMarker(const QString &markerName, int markerId)
{
    QJsonObject message { { sp::action,     sp::removePlayersListMarker },
                          { sp::markerName, markerName },
                          { sp::markerId,   markerId } };
    sendRequest(message, ProcessingTarget::Registration,
                tr("Entfernen aller „%1“-Markierungen").arg(markerName));
}

// Draw changes
// ============

void Client::requestUpdateDraw(const QString &name, const Draw::Seat &seat, int marker)
{
    QJsonObject message { { sp::action,             sp::updateDraw },
                          { sp::playersName,        name },
                          { sp::playersTableNumber, seat.table },
                          { sp::playersPairNumber,  seat.pair },
                          { sp::markerId,           marker } };
    const QString text = seat.isEmpty() ? tr("Löschen der Auslosung für „%1“").arg(name)
                                        : tr("Speichern der Auslosung für „%1“").arg(name);
    sendRequest(message, ProcessingTarget::Registration, text, name);
}

void Client::requestSetListDraw(const QVector<QString> &names,
                                const QVector<Draw::Seat> &draw)
{
    QJsonArray namesList;
    QJsonArray tables;
    QJsonArray pairs;
    for (int i = 0; i < names.count(); i++) {
        namesList.append(names.at(i));
        tables.append(draw.at(i).table);
        pairs.append(draw.at(i).pair);
    }

    QJsonObject message { { sp::action,          sp::setListDraw },
                          { sp::playersList,     namesList },
                          { sp::tablesList,      tables },
                          { sp::pairNumbersList, pairs } };

    sendRequest(message, ProcessingTarget::Registration, tr("Setzen der Auslosung aller Plätze"));
}

void Client::requestDeleteAllDraws()
{
    QJsonObject message { { sp::action, sp::deleteAllDraws } };
    sendRequest(message, ProcessingTarget::Registration, tr("Löschen aller Auslosungen"));
}

// Markers definition changes
// ==========================

void Client::requestAddMarker(const QString &name, const QColor &color,
                              Markers::MarkerSorting sorting)
{
    QJsonObject message { { sp::action,        sp::addMarker },
                          { sp::markerName,    name },
                          { sp::markerColor,   colorToString(color) },
                          { sp::markerSorting, sorting } };
    sendRequest(message, ProcessingTarget::Markers,
                tr("Hinzufügen der Markierung „%1“").arg(name));
}

void Client::requestDeleteMarker(const QString &name, int id, int sequence)
{
    QJsonObject message { { sp::action,         sp::deleteMarker },
                          { sp::markerName,     name },
                          { sp::markerId,       id },
                          { sp::markerSequence, sequence } };
    sendRequest(message, ProcessingTarget::Markers,
                tr("Frage Löschen der Markierung „%1“").arg(name));
}

void Client::requestMoveMarker(const QString &name, int id, int sequence, int direction)
{
    QJsonObject message { { sp::action,          sp::moveMarker },
                          { sp::markerName,      name },
                          { sp::markerId,        id },
                          { sp::markerSequence,  sequence },
                          { sp::markerDirection, direction } };
    sendRequest(message, ProcessingTarget::Markers,
                tr("Verschieben der Markierung „%1“ %2").arg(
                   name, direction == 1 ? tr("nach unten") : tr("nach oben")));
}

void Client::requestEditMarker(const QString &name, int id, const QString &newName,
                               const QColor &color, Markers::MarkerSorting sorting)
{
    QJsonObject message { { sp::action,        sp::editMarker },
                          { sp::markerName,    name },
                          { sp::markerId,      id },
                          { sp::newMarkerName, newName },
                          { sp::markerColor,   colorToString(color) },
                          { sp::markerSorting, sorting } };
    sendRequest(message, ProcessingTarget::Markers,
                tr("Bearbeiten der Markierung „%1“").arg(name));
}

// Special markers changes
// =======================

void Client::requestSetAbsentMarker(const QString &name, int id)
{
    QJsonObject message { { sp::action,     sp::setAbsentMarker },
                          { sp::markerName, name },
                          { sp::markerId,   id } };
    sendRequest(message, ProcessingTarget::Markers,
                id == -1 ? tr("Deaktivierung der Markierung für abwesende Anmeldungen")
                         : tr("Setzen der Markierung für abwesende Anmeldungen auf „%1“").arg(
                              name));
}

void Client::requestSinglesManagementChange(const Markers::SinglesManagement &data)
{
    QJsonObject message { { sp::action,            sp::singlesManagementChange },
                          { sp::singlesManagement, JsonHelper::singlesManagementToJson(data) } };
    sendRequest(message, ProcessingTarget::Markers,
                tr("Änderung der Einstellungen für das Behandlen allein gekommener Spieler"));
}

// Finishing the registration
// ==========================

void Client::requestFinishRegistration()
{
    QJsonObject message { { sp::action, sp::finishRegistration } };
    sendRequest(message, ProcessingTarget::Registration,
                tr("Beenden der Anmeldung"));
}

// Ranking comparison
// ==================

void Client::requestRanking()
{
    QJsonObject message { { sp::action, sp::rankingRequest } };
    sendRequest(message, ProcessingTarget::None,
                tr("Rangliste"));
}

// Ranking comparison
// ==================

void Client::requestScore(int round)
{
    QJsonObject message { { sp::action, sp::scoreRequest },
                          { sp::round,  round } };
    sendRequest(message, ProcessingTarget::None,
                tr("Ergebnisse für Runde %1").arg(round));
}

// ====================================
// Change announcements from the server
// ====================================

void Client::dataReceived(const QString &action, const QJsonObject &jsonObject)
{
    emit rx();

    if (action == sp::handshake) {
        if (! m_handshakePending) {
            return;
        }
        processHandshake(jsonObject);
        return;
    }

    const int clientId = jsonObject.value(sp::clientId).toInt();
    if (clientId == m_id) {
        stopProcessingRequest();
    } else {
        QApplication::setOverrideCursor(Qt::BusyCursor);
    }

    // Check if the request has been accepted
    if (action == sp::actionRejected) {
        QApplication::restoreOverrideCursor();
        emit statusUpdate(tr("Anfrage vom Server zurückgewiesen"));
        emit requestRejected(
            static_cast<sp::RejectReason>(jsonObject.value(sp::rejectReason).toInt()));
        return;
    }

    // Pass the data to the respective handler function
    if (m_requestHandlers.contains(action)) {
        (this->*m_requestHandlers.value(action))(jsonObject, clientId);
    } else {
        // This should not happen
        emit connectionError(tr("Die Antwort des Servers konnte nicht verarbeitet werden."));
    }

    QApplication::restoreOverrideCursor();
    verifyChecksum(jsonObject.value(sp::checksum).toString());
}

// Registration changes
// ====================

void Client::processRegisterPlayers(const QJsonObject &jsonObject, int clientId)
{
    const QString name = jsonObject.value(sp::playersName).toString();
    const int marker   = jsonObject.value(sp::markerId).toInt();
    emit message(tr("Melde „%1“ an").arg(name));

    if (! m_db->registerPlayers(name, marker)) {
       return;
    }
    statusMessage(tr("%1 „%2“ angemeldet").arg(
                     m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"), name),
                  clientId);

    m_registrationPage->processNetworkChange(clientId);
}

void Client::processRenamePlayers(const QJsonObject &jsonObject, int clientId)
{
    const QString name       = jsonObject.value(sp::playersName).toString();
    const QString newName    = jsonObject.value(sp::newPlayersName).toString();
    const int markerId       = jsonObject.value(sp::markerId).toInt();
    const QString markerName = jsonObject.value(sp::markerName).toString();

    QString markerChangeStart;
    QString markerChangeFinished;
    if (markerId != -1) {
        if (markerId == 0) {
            markerChangeStart = tr(" und entferne die Markierung");
            markerChangeFinished = tr(" und die Markierung entfernt");
        } else {
            markerChangeStart = tr(" und setze die Markierung auf „%1“").arg(markerName);
            markerChangeFinished = tr(" und die Markierung auf „%1“ gesetzt").arg(markerName);
        }
    }

    emit message(tr("Benenne „%1“ zu „%2“ um%3").arg(name, newName, markerChangeStart));

    if (! m_db->renamePlayers(name, newName, markerId)) {
        return;
    }

    statusMessage(tr("%1 „%2“ zu „%3“ umbenannt%4").arg(
                     m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                     name, newName, markerChangeFinished),
                  clientId);

    m_registrationPage->processNetworkChange(clientId);
    m_registrationPage->checkUpdateDrawOverviewPage(clientId == m_id, newName);
    emit nameEdited();
}

void Client::processAssemblePair(const QJsonObject &jsonObject, int clientId)
{
    const QJsonArray namesData = jsonObject.value(sp::playersName).toArray();
    const QString player1      = namesData.at(0).toString();
    const QString player2      = namesData.at(1).toString();
    const QString separator    = namesData.at(2).toString();
    const int marker           = jsonObject.value(sp::markerId).toInt();
    emit message(tr("Füge „%1“ und „%2“ zu einem Paar zusammen").arg(player1, player2));

    if (! m_db->assemblePair(player1, player2, separator, marker)) {
        return;
    }

    statusMessage(tr("„%1“ und „%2“ zu einem Paar zusammengefügt").arg(player1, player2),
                  clientId);

    m_registrationPage->processNetworkChange(clientId);
    m_registrationPage->checkUpdateDrawOverviewPage(clientId == m_id,
                                                    player1 + separator + player2);
}

void Client::processDeletePlayers(const QJsonObject &jsonObject, int clientId)
{
    const QString name = jsonObject.value(sp::playersName).toString();
    emit message(tr("Lösche „%1“").arg(name));

    if (! m_db->deletePlayers(name)) {
        return;
    }

    statusMessage(tr("%1 „%2“ gelöscht").arg(
                     m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"), name),
                  clientId);

    m_registrationPage->processNetworkChange(clientId);
    m_registrationPage->checkUpdateDrawOverviewPage(clientId == m_id, name, true);
}

void Client::processUpdatePlayersMarker(const QJsonObject &jsonObject, int clientId)
{
    const QString name       = jsonObject.value(sp::playersName).toString();
    const QString markerName = jsonObject.value(sp::markerName).toString();
    const int markerId       = jsonObject.value(sp::markerId).toInt();

    if (markerId != 0) {
        emit message(tr("Markiere „%1“ als „%2“").arg(name, markerName));
    } else {
        emit message(tr("Entferne Markierung von „%1“").arg(name));
    }

    if (! m_db->updateMarker(name, markerId)) {
        return;
    }

    if (markerId != 0) {
        statusMessage(tr("%1 „%2“ als „%3“ markiert").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                         name, markerName),
                      clientId);
    } else {
        statusMessage(tr("Markierung von %1 „%2“ entfernt").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"), name),
                      clientId);
    }

    m_registrationPage->processNetworkChange(clientId);
}

void Client::processDeleteMarkedPlayers(const QJsonObject &jsonObject, int clientId)
{
    const QString markerName = jsonObject.value(sp::markerName).toString();
    const int markerId       = jsonObject.value(sp::markerId).toInt();
    emit message(tr("Lösche alle als „%1“ markierten %2").arg(
                    markerName, m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    if (! m_db->deleteMarkedPlayers(markerId)) {
        return;
    }

    statusMessage(tr("Alle als „%1“ markierten %2 gelöscht").arg(
                     markerName, m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")),
                  clientId);

    m_registrationPage->processNetworkChange(clientId);
    m_registrationPage->checkUpdateDrawOverviewPage(clientId == m_id);
}

void Client::processDeleteUnmarkedPlayers(const QJsonObject &, int clientId)
{
    emit message(tr("Lösche alle unmarkierten %1").arg(
                    m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    if (! m_db->deleteUnmarkedPlayers()) {
        return;
    }

    statusMessage(tr("Alle unmarkierten %1 gelöscht").arg(
                     m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")),
                  clientId);

    m_registrationPage->processNetworkChange(clientId);
    m_registrationPage->checkUpdateDrawOverviewPage(clientId == m_id);
}

void Client::processDeleteAllPlayers(const QJsonObject &, int clientId)
{
    emit message(tr("Lösche alle %1").arg(m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    if (! m_db->deleteAllPlayers()) {
        return;
    }

    statusMessage(tr("Alle %1 gelöscht").arg(
                     m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")),
                  clientId);

    m_registrationPage->processNetworkChange(clientId);
    m_registrationPage->checkUpdateDrawOverviewPage(clientId == m_id);
}

void Client::processUpdateListMarker(const QJsonObject &jsonObject, int clientId)
{
    const QString markerName = jsonObject.value(sp::markerName).toString();
    const int markerId       = jsonObject.value(sp::markerId).toInt();

    if (markerId != 0) {
        emit message(tr("Markiere alle %1 als „%2“").arg(
                        m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                        markerName));
    } else {
        emit message(tr("Entferne alle Markierungen"));
    }

    if (! m_db->listUpdateMarker(markerId)) {
        return;
    }

    if (markerId != 0) {
        statusMessage(tr("Alle %1 als „%2“ markiert").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                         markerName),
                      clientId);
    } else {
        statusMessage(tr("Alle Markierungen entfernt"), clientId);
    }

    m_registrationPage->processNetworkChange(clientId);
}

void Client::processRemoveListMarker(const QJsonObject &jsonObject, int clientId)
{
    const QString markerName = jsonObject.value(sp::markerName).toString();
    const int markerId       = jsonObject.value(sp::markerId).toInt();
    emit message(tr("Entferne alle „%1“-Markierungen").arg(markerName));

    if (! m_db->listRemoveMarker(markerId)) {
        return;
    }

    statusMessage(tr("Alle „%1“-Markierungen entfernt").arg(markerName), clientId);

    m_registrationPage->processNetworkChange(clientId);
}

// Draw changes
// ============

void Client::processUpdateDraw(const QJsonObject &jsonObject, int clientId)
{
    const QString name    = jsonObject.value(sp::playersName).toString();
    const Draw::Seat seat = { jsonObject.value(sp::playersTableNumber).toInt(),
                              jsonObject.value(sp::playersPairNumber).toInt() };
    if (seat.isEmpty()) {
        emit message(tr("Lösche Auslosung für „%1“").arg(name));
    } else {
        emit message(tr("Speichere Auslosung für „%1“").arg(name));
    }

    if (! m_db->updateDraw(name, seat, jsonObject.value(sp::markerId).toInt())) {
        return;
    }

    if (seat.isEmpty()) {
        statusMessage(tr("Auslosung für „%1“ gelöscht").arg(name), clientId);
    } else {
        statusMessage(tr("Auslosung für „%1“ gespeichert (Tisch %2 Paar %3)").arg(
                         name, QString::number(seat.table), QString::number(seat.pair)), clientId);
    }

    m_registrationPage->processNetworkChange(clientId);
    m_registrationPage->checkUpdateDrawOverviewPage(clientId == m_id, name, seat.isEmpty());
}

void Client::processSetListDraw(const QJsonObject &jsonObject, int clientId)
{
    emit message(tr("Setze Auslosung aller Plätze"));

    QVector<QString> names;
    QVector<Draw::Seat> draw;
    const QJsonArray namesData  = jsonObject.value(sp::playersList).toArray();
    const QJsonArray tablesData = jsonObject.value(sp::tablesList).toArray();
    const QJsonArray pairsData  = jsonObject.value(sp::pairNumbersList).toArray();
    for (int i = 0; i < namesData.count(); i++) {
        names.append(namesData.at(i).toString());
        draw.append(Draw::Seat { tablesData.at(i).toInt(), pairsData.at(i).toInt() });
    }

    if (! m_db->setListDraw(names, draw)) {
        return;
    }

    statusMessage(tr("Auslosung aller Plätze gesetzt"), clientId);
    m_registrationPage->processNetworkChange(clientId);
    m_registrationPage->checkUpdateDrawOverviewPage(clientId == m_id);
}

void Client::processDeleteAllDraws(const QJsonObject &, int clientId)
{
    emit message(tr("Lösche alle Auslosungen"));

    if (! m_db->deleteAllDraws()) {
        return;
    }

    statusMessage(tr("Alle Auslosungen gelöscht"), clientId);

    m_registrationPage->processNetworkChange(clientId);
    m_registrationPage->checkUpdateDrawOverviewPage(clientId == m_id);
}

// Marker definition changes
// =========================

void Client::processAddMarker(const QJsonObject &jsonObject, int clientId)
{
    const QString name = jsonObject.value(sp::markerName).toString();
    emit message(tr("Füge Markierung „%1“ hinzu").arg(name));

    if (! m_db->addMarker(name, colorFromValue(jsonObject.value(sp::markerColor)),
                          static_cast<Markers::MarkerSorting>(
                              jsonObject.value(sp::markerSorting).toInt()))) {
        return;
    }

    statusMessage(tr("Markierung „%1“ hinzugefügt").arg(name), clientId);

    m_markersWidget->processNetworkChange();
}

void Client::processDeleteMarker(const QJsonObject &jsonObject, int clientId)
{
    const QString name = jsonObject.value(sp::markerName).toString();
    emit message(tr("Lösche Markierung „%1“").arg(name));

    if (! m_db->deleteMarker(jsonObject.value(sp::markerId).toInt(),
                             jsonObject.value(sp::markerSequence).toInt())) {
        return;
    }

    statusMessage(tr("Markierung „%1“ gelöscht").arg(name), clientId);

    m_markersWidget->processNetworkChange();
}

void Client::processMoveMarker(const QJsonObject &jsonObject, int clientId)
{
    const QString name  = jsonObject.value(sp::markerName).toString();
    const int direction = jsonObject.value(sp::markerDirection).toInt();
    emit message(tr("Verschiebe Markierung „%1“ %2").arg(
                    name, direction == 1 ? tr("nach unten") : tr("nach oben")));

    if (! m_db->moveMarker(jsonObject.value(sp::markerId).toInt(),
                           jsonObject.value(sp::markerSequence).toInt(),
                           direction)) {
        return;
    }

    statusMessage(tr("Markierung „%1“ %2 verschoben").arg(
                     name, direction == 1 ? tr("nach unten") : tr("nach oben")),
                  clientId);

    m_markersWidget->processNetworkChange();
}

void Client::processEditMarker(const QJsonObject &jsonObject, int clientId)
{
    const QString name = jsonObject.value(sp::markerName).toString();
    emit message(tr("Speichere Änderungen an Markierung „%1“").arg(name));

    if (! m_db->editMarker(jsonObject.value(sp::markerId).toInt(),
                           jsonObject.value(sp::newMarkerName).toString(),
                           colorFromValue(jsonObject.value(sp::markerColor)),
                           static_cast<Markers::MarkerSorting>(
                               jsonObject.value(sp::markerSorting).toInt()))) {
        return;
    }

    statusMessage(tr("Änderungen an Markierung „%1“ gespeichert").arg(name), clientId);

    m_markersWidget->processNetworkChange();
}

// Special markers changes
// =======================

void Client::processSetAbsentMarker(const QJsonObject &jsonObject, int clientId)
{
    const QString name = jsonObject.value(sp::markerName).toString();
    const int markerId = jsonObject.value(sp::markerId).toInt();

    emit message(markerId == -1 ? tr("Deaktiviere Markierung für abwesende Anmeldungen")
                                : tr("Setze Markierung für abwesende Anmeldungen auf „%1“").arg(
                                     name));

    m_markersWidget->setCheckableMarker(Markers::Absent, markerId);
    m_tournamentSettings->setMarker(Markers::Absent, markerId);

    statusMessage(markerId == -1 ? tr("Markierung für abwesende Anmeldungen deaktiviert")
                                 : tr("Markierung für abwesende Anmeldungen auf „%1“ gesetzt").arg(
                                      name),
                  clientId);

    m_markersWidget->processNetworkChange();

    if (markerId != -1) {
        m_markersWidget->warnAboutSpecialMarkers();
    }
}

void Client::processSinglesManagementChange(const QJsonObject &jsonObject, int clientId)
{
    emit message(tr("Übernehme Änderungen an den Einstellungen für das Behandlen allein gekommener "
                    "Spieler"));

    const Markers::SinglesManagement data = JsonHelper::parseSinglesManagement(
        jsonObject.value(sp::singlesManagement).toObject());
    m_markersWidget->setSinglesManagement(data);
    m_tournamentSettings->setSinglesManagement(data);
    statusMessage(tr("Änderungen an den Einstellungen für das Behandeln allein gekommener Spieler "
                     "übernommen"),
                  clientId);

    m_markersWidget->processNetworkChange();

    if (data.enabled) {
        m_markersWidget->warnAboutSpecialMarkers();
    }
}

// Finishing the registration
// ==========================

void Client::processFinishRegistration(const QJsonObject &, int clientId)
{
    emit message(tr("Beende die Anmeldung"));
    m_registrationPage->networkFinishRegistration(clientId);
    statusMessage(tr("Anmeldung beendet"), clientId);
}

// Ranking comparison
// ==================

void Client::processRankingRequest(const QJsonObject &jsonObject, int)
{
    QVector<Database::RankData> ranking;

    if (! jsonObject.value(sp::rankingAvailable).toBool()) {
        emit statusUpdate(tr("Antwort der Ranglistenanfrage empfangen"));
        emit message(tr("Es ist noch keine Rangliste verfügbar"));
        emit compareRanking(ranking);
        return;
    }

    const QJsonArray names         = jsonObject.value(sp::playersList).toArray();
    const QJsonArray boogers       = jsonObject.value(sp::boogersList).toArray();
    const QJsonArray goals         = jsonObject.value(sp::goalsList).toArray();
    const QJsonArray opponentGoals = jsonObject.value(sp::opponentGoalsList).toArray();
    const QJsonArray playedBoogers = jsonObject.value(sp::playedBoogersList).toArray();

    for (int i = 0; i < names.count(); i++) {
        Database::RankData rank;
        rank.name = names.at(i).toString();
        rank.boogers = boogers.at(i).toInt();
        rank.goals = goals.at(i).toInt();
        rank.opponentGoals = opponentGoals.at(i).toInt();
        rank.playedBoogers = playedBoogers.at(i).toInt();
        ranking.append(rank);
    }

    emit message(tr("Rangliste empfangen"));
    emit statusUpdate(tr("Rangliste empfangen"));
    emit compareRanking(ranking);
}

// Score change
// ============

void Client::processScoreChange(const QJsonObject &jsonObject, int)
{
    const int round = jsonObject.value(sp::round).toInt();
    emit message(tr("Änderung der Ergebnisse für Runde %1 am Server").arg(round));
    emit remoteChecksumChanged(round, jsonObject.value(sp::scoreChecksum).toString());
}

// Score comparison
// ================

void Client::processScoreRequest(const QJsonObject &jsonObject, int)
{
    Scores::Score score;

    const auto jsonTables = jsonObject.value(sp::tablesList).toArray();
    for (const auto &table : jsonTables) {
        score.tables.append(table.toInt());
    }

    const auto jsonPairs = jsonObject.value(sp::playersList).toArray();
    for (const auto &pair : jsonPairs) {
        score.pairs.append(pair.toString());
    }

    const auto jsonScores = jsonObject.value(sp::scoresList).toArray();
    for (const auto &jsonPairResult : jsonScores) {
        const auto jsonPairResultArray = jsonPairResult.toArray();
        QVector<int> pairResult;
        for (const auto &result : jsonPairResultArray) {
            pairResult.append(result.toInt());
        }
        score.scores.append(pairResult);
    }

    emit message(tr("Rundenergebnisse empfangen"));
    emit statusUpdate(tr("Rundenergebnisse empfangen"));
    emit compareScore(score);
}
