/* SPDX-FileCopyrightText: 2018-2020 Tobias Leupold <tobias.leupold@gmx.de>

   SPDX-License-Identifier: GPL-3.0-or-later
*/

#ifndef CLIENT_H
#define CLIENT_H

// Local includes

#include "network/AbstractNetworkEngine.h"
#include "network/ServerProtocol.h"

#include "shared/Database.h"
#include "shared/Markers.h"
#include "shared/Draw.h"
#include "shared/Scores.h"

// Qt includes
#include <QAbstractSocket>

// Local classes
class SharedObjects;
class Settings;
class TournamentSettings;
class StringTools;
class RegistrationPage;
class MarkersWidget;

// Qt classes
class QHostAddress;
class QTimer;
class QJsonObject;

class Client : public AbstractNetworkEngine
{
    Q_OBJECT

public:
    enum ProcessingTarget {
        None,
        Registration,
        Markers
    };

    explicit Client(QObject *parent, int connectTimeout, int requestTimeout,
                    SharedObjects *sharedObjects, RegistrationPage *registrationPage);
    void setChecksum(const QString *checksum);
    void connectToServer(const QHostAddress &address, quint16 port);
    void closeConnection();
    QString serverIp() const;
    int serverPort() const;
    int id() const;
    bool connected() const;
    void abortHandshake();
    void checkAbortHandshake();

    // Change request functions
    // ========================

    // Registration changes

    void requestRegisterPlayers(const QString &name, int marker);
    void requestRenamePlayers(const QString &name, const QString &newName,
                              int markerId, const QString &markerName);
    void requestAssemblePair(const QString &player1, const QString &player2,
                             const QString &separator, int assignedMarker);
    void requestDeletePlayers(const QString &name);
    void requestUpdateMarker(const QString &name, const QString &markerName, int markerId);

    void requestDeleteMarkedPlayers(const QString &markerName, int markerId);
    void requestDeleteUnmarkedPlayers();
    void requestDeleteAllPlayers();
    void requestListUpdateMarker(const QString &markerName, int markerId);
    void requestListRemoveMarker(const QString &markerName, int markerId);

    // Draw changes
    void requestUpdateDraw(const QString &name, const Draw::Seat &seat, int Marker);
    void requestSetListDraw(const QVector<QString> &names, const QVector<Draw::Seat> &draw);
    void requestDeleteAllDraws();

    // Markers definition changes
    void requestAddMarker(const QString &name, const QColor &color,
                          Markers::MarkerSorting sorting);
    void requestDeleteMarker(const QString &name, int id, int sequence);
    void requestMoveMarker(const QString &name, int id, int sequence, int direction);
    void requestEditMarker(const QString &name, int id, const QString &newName,
                           const QColor &color, Markers::MarkerSorting sorting);

    // Special markers changes
    void requestSetAbsentMarker(const QString &name, int id);
    void requestSinglesManagementChange(const Markers::SinglesManagement &data);

    // Finishing the registration
    void requestFinishRegistration();

public slots:
    // Ranking comparison
    void requestRanking();

    // Score comparison
    void requestScore(int round);

signals:
    void tx();
    void rx();

    void message(const QString &text);
    void connectResult(bool success, const QString &error = QString());
    void connectionError(const QString &text);
    void requestRejected(ServerProtocol::RejectReason reason);
    void statusUpdate(const QString &text);
    void namesDiffer(QStringList &localNames, QStringList &remoteNames);
    void nameEdited();
    void remoteChecksumChanged(int round, const QString &checksum);

    void compareRanking(const QVector<Database::RankData> &ranking);
    void compareScore(Scores::Score &score);

private slots:
    void socketConnected();
    void error(QAbstractSocket::SocketError);
    void connectionTimeout();
    void requestTimeout();

private: // Functions
    void statusMessage(const QString &message, int clientId);
    void sendRequest(QJsonObject &jsonObject, ProcessingTarget target,
                     const QString &request, const QString &scrollToTarget = QString());
    void processHandshake(const QJsonObject &jsonObject);
    void verifyChecksum(const QString &checksum);
    void stopProcessingRequest();

    void dataReceived(const QString &action, const QJsonObject &jsonObject) override;

    // Server change announcement handlers
    // ===================================

    // Helper macro to keep the header a bit more tidy
    #define CHANGE_HANDLER(name) void process##name(const QJsonObject &jsonObject, int clientId)

    // Registration changes

    CHANGE_HANDLER(RegisterPlayers);
    CHANGE_HANDLER(RenamePlayers);
    CHANGE_HANDLER(AssemblePair);
    CHANGE_HANDLER(DeletePlayers);
    CHANGE_HANDLER(UpdatePlayersMarker);

    CHANGE_HANDLER(DeleteMarkedPlayers);
    CHANGE_HANDLER(DeleteUnmarkedPlayers);
    CHANGE_HANDLER(DeleteAllPlayers);
    CHANGE_HANDLER(UpdateListMarker);
    CHANGE_HANDLER(RemoveListMarker);

    // Draw changes
    CHANGE_HANDLER(UpdateDraw);
    CHANGE_HANDLER(SetListDraw);
    CHANGE_HANDLER(DeleteAllDraws);

    // Markers definition changes
    CHANGE_HANDLER(AddMarker);
    CHANGE_HANDLER(DeleteMarker);
    CHANGE_HANDLER(MoveMarker);
    CHANGE_HANDLER(EditMarker);

    // Special markers changes
    CHANGE_HANDLER(SetAbsentMarker);
    CHANGE_HANDLER(SinglesManagementChange);

    // Finishing the registration
    CHANGE_HANDLER(FinishRegistration);

    // Ranking comparison
    CHANGE_HANDLER(RankingRequest);

    // Score change
    CHANGE_HANDLER(ScoreChange);

    // Score comparison
    CHANGE_HANDLER(ScoreRequest);

    #undef CHANGE_HANDLER

private: // Variables
    const int m_connectTimeout;
    Database *m_db;
    Settings *m_settings;
    TournamentSettings *m_tournamentSettings;
    StringTools *m_stringTools;
    RegistrationPage *m_registrationPage;
    MarkersWidget *m_markersWidget;
    const QHash<QString, void(Client::*)(const QJsonObject &, int)> m_requestHandlers;
    const QString *m_checksum;

    int m_id;
    bool m_connected = false;
    bool m_connectionPending = false;
    bool m_handshakePending = false;
    QTimer *m_connectionTimer;
    QTimer *m_requestTimer;
    QVector<ProcessingTarget> m_processingTarget;

};

#endif // CLIENT_H
