/* SPDX-FileCopyrightText: 2020 Tobias Leupold <tobias.leupold@gmx.de>

   SPDX-License-Identifier: GPL-3.0-or-later
*/

// Local includes

#include "CompareScoreDialog.h"

#include "ClientPage/Client.h"

#include "shared/SharedObjects.h"
#include "shared/Database.h"
#include "shared/TournamentSettings.h"
#include "shared/StringTools.h"
#include "shared/SharedStyles.h"
#include "shared/TableDelegate.h"

// Qt includes
#include <QDebug>
#include <QApplication>
#include <QMessageBox>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QLabel>
#include <QTimer>
#include <QSet>
#include <QTableWidget>
#include <QHeaderView>
#include <QScrollBar>
#include <QMultiMap>

// C++ includes
#include <utility>
#include <algorithm>

CompareScoreDialog::CompareScoreDialog(QWidget *parent, SharedObjects *sharedObjects,
                                       Client *client, const int *currentRound)
    : TitleDialog(parent),
      m_db(sharedObjects->database()),
      m_tournamentSettings(sharedObjects->tournamentSettings()),
      m_stringTools(sharedObjects->stringTools()),
      m_client(client),
      m_currentRound(currentRound)
{
    setTitle(tr("Vergleich von Ergebnissen"));

    auto box = new QGroupBox;
    box->setProperty("notitle", true);
    auto *boxLayout = new QVBoxLayout(box);
    layout()->addWidget(box);

    m_requesting = new QLabel(tr("<i>Anfrage läuft …</i>"));
    m_requesting->setAlignment(Qt::AlignCenter);
    m_requesting->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Expanding);
    boxLayout->addWidget(m_requesting);

    m_differencesLabel = new QLabel(tr(
        "<span style=\"%1\">Bei folgenden Paaren gibt es Unterschiede zwischen dem Datensatz "
        "des Servers und dem Lokalen:</span>").arg(SharedStyles::redText));
    m_differencesLabel->setWordWrap(true);
    boxLayout->addWidget(m_differencesLabel);

    m_noDifferencesLabel = new QLabel(tr(
        "<span style=\"%1\">Die Ergebnisse des Servers sind identisch mit den lokal eingegebenen."
        "</span>").arg(SharedStyles::greenText));
    m_noDifferencesLabel->setWordWrap(true);
    boxLayout->addWidget(m_noDifferencesLabel);

    m_differences = new QTableWidget;
    m_differences->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_differences->setItemDelegate(new TableDelegate(this));
    m_differences->verticalHeader()->hide();
    m_differences->setShowGrid(false);
    m_differences->setWordWrap(false);
    m_differences->setSelectionMode(QTableWidget::NoSelection);
    m_differences->setFocusPolicy(Qt::NoFocus);
    m_differences->setContextMenuPolicy(Qt::CustomContextMenu);
    boxLayout->addWidget(m_differences);

    m_differences->setColumnCount(5);
    m_differences->setHorizontalHeaderLabels({ tr("Tisch"), tr("Paar"), tr("Bobbl"),
                                               tr("Server"), tr("Lokal") });

    m_dataChangedLabel = new QLabel(tr(
        "<span style=\"%1\">Seit dem letzten Abruf wurde entweder eine andere Runde geöffnet, "
        "oder der Datenbestand hat sich lokal oder auf dem Server geändert. Bitte die "
        "Server-Ergebnisse erneut abfragen!</span>").arg(
        SharedStyles::redText));
    m_dataChangedLabel->setWordWrap(true);
    boxLayout->addWidget(m_dataChangedLabel);

    addButtonBox();

    m_request = buttonBox()->addButton(QDialogButtonBox::Yes);
    m_request->setText(tr("Server-Ergebnisse abfragen"));
    connect(m_request, &QPushButton::clicked, this, &CompareScoreDialog::requestScore);

    m_close = buttonBox()->addButton(QDialogButtonBox::Close);
    connect(m_close, &QPushButton::clicked, this, &QDialog::reject);

    QTimer::singleShot(0, this, &CompareScoreDialog::requestScore);
}

void CompareScoreDialog::clientDisconnected()
{
    QApplication::restoreOverrideCursor();
    QMessageBox::critical(this, windowTitle(),
                          tr("Die Netzwerkverbindung wurde beendet. Der Dialog wird geschlossen."));
    QDialog::reject();
}

void CompareScoreDialog::remoteChecksumChanged(int round)
{
    if (round == (*m_currentRound)) {
        dataChanged();
    }
}

void CompareScoreDialog::dataChanged()
{
    m_dataChangedLabel->show();
    m_request->setEnabled(true);
    m_request->setFocus();

    resizeToContents();
    if (m_differences->isVisible()) {
        resizeToTableSize();
    }
}

void CompareScoreDialog::requestScore()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    m_request->setEnabled(false);

    m_requesting->show();
    m_differencesLabel->hide();
    m_noDifferencesLabel->hide();
    m_differences->hide();
    m_dataChangedLabel->hide();

    emit scoreRequested(*m_currentRound);
}

void CompareScoreDialog::compareScore(Scores::Score &remoteScore)
{
    QApplication::restoreOverrideCursor();

    m_dataChangedLabel->hide();
    m_requesting->hide();

    setTitle(tr("Vergleich der Ergebnisse für Runde %1").arg(*m_currentRound));

    // Get the local score of the openend round
    const auto [ localScore, success ] = m_db->scores(*m_currentRound);
    if (! success) {
        QDialog::reject();
        return;
    }

    // Generate a sorted list of all pair names
    QSet<QString> allPairsSet;
    for (const auto &pair : localScore.pairs) {
        allPairsSet.insert(pair);
    }
    for (const auto &pair : remoteScore.pairs) {
        allPairsSet.insert(pair);
    }
#if QT_VERSION >= QT_VERSION_CHECK(5, 14, 0)
    QVector<QString> allPairs(allPairsSet.begin(), allPairsSet.end());
#else
    auto allPairs = allPairsSet.toList();
#endif
    m_stringTools->sort(allPairs);

    // Find differences

    QVector<Id> differing;
    QVector<int> missingRemotely;
    QVector<int> missingLocally;

    for (const auto &pair : allPairs) {
        const int remoteId = remoteScore.pairs.contains(pair) ? remoteScore.pairs.indexOf(pair)
                                                              : -1;
        const int localId = localScore.pairs.contains(pair) ? localScore.pairs.indexOf(pair) : -1;

        if (remoteId != -1 && localId != -1) {
            if (remoteScore.scores.at(remoteId) != localScore.scores.at(localId)) {
                differing.append({ remoteId, localId });
            }
        } else if (remoteId == -1) {
            missingRemotely.append(localId);
        } else if (localId == -1) {
            missingLocally.append(remoteId);
        }
    }

    // Display the differnces

    m_differences->show();
    m_differences->clearContents();
    m_differences->setRowCount(differing.count() + missingRemotely.count()
                               + missingLocally.count());
    m_differences->setColumnHidden(0, m_tournamentSettings->noTableNumbers());

    int row = 0;

    // Differing results

    QMultiMap<int, Id> differingSorted;
    for (const auto &id : differing) {
        differingSorted.insert(localScore.tables.at(id.local / 2), id);
    }

    QMapIterator<int, Id> it1(differingSorted);
    while (it1.hasNext()) {
        it1.next();

        const auto &id = it1.value();
        const int localTable = it1.key();
        const int remoteTable = remoteScore.tables.at(id.remote / 2);

        m_differences->setItem(row, 0, new QTableWidgetItem(localTable == remoteTable
            ? QString::number(localTable)
            : tr("Lokal: %1\nServer: %2").arg(QString::number(localTable),
                                              QString::number(remoteTable))));

        m_differences->setItem(row, 1, new QTableWidgetItem(localScore.pairs.at(id.local)));

        QString boogers;
        QString remote;
        QString local;
        const auto &remoteResult = remoteScore.scores.at(id.remote);
        const auto &localResult = localScore.scores.at(id.local);
        for (int booger = 0; booger < remoteResult.count(); booger++) {
            if (remoteResult.at(booger) != localResult.at(booger)) {
                boogers.append(tr("%1.\n").arg(booger + 1));
                remote.append(tr("%1\n").arg(m_stringTools->boogify(remoteResult.at(booger))));
                local.append(tr("%1\n").arg(m_stringTools->boogify(localResult.at(booger))));
            }
        }
        m_differences->setItem(row, 2, new QTableWidgetItem(boogers.trimmed()));
        m_differences->setItem(row, 3, new QTableWidgetItem(remote.trimmed()));
        m_differences->setItem(row, 4, new QTableWidgetItem(local.trimmed()));

        row++;
    }

    // Results missing on the server
    QMultiMap<int, int> remoteMissingTables;
    for (const int id : missingRemotely) {
        remoteMissingTables.insert(localScore.tables.at(id / 2), id);
    }
    QMapIterator<int, int> it2(remoteMissingTables);
    while (it2.hasNext()) {
        it2.next();
        m_differences->setItem(row, 0, new QTableWidgetItem(tr("Lokal: %1").arg(it2.key())));
        m_differences->setItem(row, 1, new QTableWidgetItem(localScore.pairs.at(it2.value())));
        m_differences->setItem(row, 3, new QTableWidgetItem(tr("Fehlt")));
        row++;
    }

    // Results missing locally
    QMultiMap<int, int> localMissingTables;
    for (const int id : missingLocally) {
        localMissingTables.insert(remoteScore.tables.at(id / 2), id);
    }
    QMapIterator<int, int> it3(localMissingTables);
    while (it3.hasNext()) {
        it3.next();
        m_differences->setItem(row, 0, new QTableWidgetItem(tr("Server: %1").arg(it3.key())));
        m_differences->setItem(row, 1, new QTableWidgetItem(remoteScore.pairs.at(it3.value())));
        m_differences->setItem(row, 4, new QTableWidgetItem(tr("Fehlt")));
        row++;
    }

    m_differences->resizeRowsToContents();
    m_differences->resizeColumnsToContents();

    const bool identical = m_differences->rowCount() == 0;
    m_differencesLabel->setVisible(! identical);
    m_noDifferencesLabel->setVisible(identical);
    m_differences->setVisible(! identical);

    if (identical) {
        resizeToContents();
        m_close->setFocus();
    } else {
        resizeToTableSize();
        m_request->setFocus();
    }
}

void CompareScoreDialog::resizeToTableSize()
{
    QApplication::processEvents();
    QTimer::singleShot(0, [this]
    {
        const auto *parentWidget = qobject_cast<QWidget *>(parent());
        const auto parentSize = parentWidget->size();

        // FIXME: If anybody ever may know how to get the correct scroll bar
        //        width/height _before_ it's actually shown, tell me :-P
        const int scrollBarWidth = std::min(m_differences->verticalScrollBar()->width(),
                                            m_differences->horizontalScrollBar()->height());

        // Step 1: Adjust the width so that all columns are shown if the parent's
        // width allows this (we don't become wider than the parent)

        const int remainingDialogWidth = width() - m_differences->width();
        const int maximumTableWidth = m_differences->horizontalHeader()->length()
                                      + m_differences->width() - m_differences->viewport()->width()
                                      + scrollBarWidth;

        resize(std::min(parentSize.width(), maximumTableWidth + remainingDialogWidth), height());

        // Step 2: Adjust the width so that all rows are shown if the parent's
        // height allows this (we don't become taller than the parent)

        const int remainingDialogHeight = height() - m_differences->height();

        int tableHeight = 0;
        for (int i = 0; i < m_differences->rowCount(); i++) {
            tableHeight += m_differences->rowHeight(i);
        }
        tableHeight += m_differences->height() - m_differences->viewport()->height();

        resize(width(), std::min(tableHeight + remainingDialogHeight, parentSize.height()));

        // Step 3: (Somehow) Center this dialog on the parent
        move(parentWidget->geometry().center() - rect().center());
    });
}
