/* SPDX-FileCopyrightText: 2010-2020 Tobias Leupold <tobias.leupold@gmx.de>

   SPDX-License-Identifier: GPL-3.0-or-later
*/

// Local includes

#include "ExportTournamentDialog.h"
#include "BasicExportPage.h"
#include "ExportDatasetsPage.h"
#include "ExportFormattingPage.h"

#include "shared/SharedObjects.h"
#include "shared/StringTools.h"
#include "shared/TournamentSettings.h"
#include "shared/Tournament.h"
#include "shared/StackedWidgetList.h"
#include "shared/Settings.h"
#include "shared/Urls.h"

#include "version.h"

// Qt includes
#include <QDebug>
#include <QGroupBox>
#include <QLabel>
#include <QVBoxLayout>
#include <QCheckBox>
#include <QLineEdit>
#include <QDir>
#include <QApplication>
#include <QTextCodec>
#include <QDateTime>
#include <QTableWidget>
#include <QRadioButton>
#include <QComboBox>

ExportTournamentDialog::ExportTournamentDialog(QWidget *parent, SharedObjects *sharedObjects)
    : FileDialog(parent, sharedObjects->database()),
      m_db(sharedObjects->database()),
      m_stringTools(sharedObjects->stringTools()),
      m_tournamentSettings(sharedObjects->tournamentSettings()),
      m_dateFormat(m_stringTools->dateFormat()),
      m_timeFormat(m_stringTools->timeFormat())
{
    setTitle(tr("Turnierdaten exportieren"));
    setOkButtonText(tr("Exportieren"));
    setFileDialogTitle(tr("Bitte einen Dateinamen für die Turnierdaten-Exportdatei auswählen"));

    auto *stack = new StackedWidgetList;

    m_basicExportPage = new BasicExportPage(QFileInfo(m_db->dbFile()).completeBaseName());
    stack->addWidget(m_basicExportPage);
    connect(m_basicExportPage, &BasicExportPage::closeAfterExportToggled,
            this, [this](bool state)
            {
                setButtonText(QDialogButtonBox::Cancel, state ? tr("Abbrechen") : tr("Schließen"));
            });
    layout()->addWidget(stack);

    m_exportDatasetsPage = new ExportDatasetsPage(m_db->allRounds());
    connect(m_exportDatasetsPage, &ExportDatasetsPage::checkExportPossible,
            this, &ExportTournamentDialog::checkExportPossible);
    stack->addWidget(m_exportDatasetsPage);

    m_exportFormattingPage = new ExportFormattingPage(sharedObjects->settings()->boogerSymbol());
    stack->addWidget(m_exportFormattingPage);

    stack->setCurrentIndex(0);
}

void ExportTournamentDialog::checkExportPossible()
{
    setOkButtonEnabled(m_exportDatasetsPage->exportGenericData()
                       || m_exportDatasetsPage->exportRanking()
                       || m_exportDatasetsPage->exportRoundResults());
}

void ExportTournamentDialog::accept()
{
    auto title = m_basicExportPage->title();
    if (! title.isEmpty()) {
        setTargetFileBaseName(title);
    } else {
        setTargetFileBaseName(tr("Protokoll"));
    }

    const auto fileFormat = m_basicExportPage->fileFormat();

    QString message;
    switch (fileFormat) {
    case BasicExportPage::FileFormat::HTML:
        setTargetFileExtension(QStringLiteral("htm"));
        setTargetFileDescription(tr("HTML-Dateien"));
        message = tr("Exportiere HTML-Daten …");
        break;
    case BasicExportPage::FileFormat::CSV:
        setTargetFileExtension(QStringLiteral("csv"));
        setTargetFileDescription(tr("CSV-Dateien"));
        message = tr("Exportiere CSV-Daten …");
        break;
    }

    QString journalFileName = getTargetFileName(m_db->dbPath());
    if (journalFileName.isEmpty()) {
        return;
    }

    QFile journalFile(journalFileName);
    journalFile.open(QIODevice::WriteOnly | QIODevice::Text);

    QTextStream outStream(&journalFile);
    outStream.setCodec(QTextCodec::codecForName("UTF-8"));
    emit updateStatus(message);
    QApplication::setOverrideCursor(Qt::WaitCursor);

    switch (fileFormat) {
    case BasicExportPage::FileFormat::HTML:
        saveHtmlFile(outStream);
        break;
    case BasicExportPage::FileFormat::CSV:
        saveCsvFile(outStream);
        break;
    }

    journalFile.close();
    QApplication::restoreOverrideCursor();
    emit updateStatus(tr("Daten in Datei „%1“ exportiert").arg(
                         QDir::toNativeSeparators(journalFileName)));

    successMessage(tr("Turnierdaten exportieren"),
                   tr("Die Daten wurden erfolgreich in die Datei „%1“ exportiert.").arg(
                      QDir::toNativeSeparators(journalFileName)));

    if (m_basicExportPage->openAfterExport()) {
        openFile(journalFileName);
    }

    if (m_basicExportPage->closeAfterExport()) {
        QDialog::accept();
    }
}

void ExportTournamentDialog::saveHtmlFile(QTextStream &outStream)
{
    // Add the HTML header
    outStream << "<!DOCTYPE html>\n\n"
              << "<html lang=\"de\">\n\n"
              << "<head>\n\n"
              << "<meta charset=\"utf-8\"/>\n\n"
              << "<title>" << targetFileBaseName().toHtmlEscaped() << "</title>\n\n"
              << "<style>\n"
              << "table { border-spacing: 0px; }\n"
              << "th, td { padding: 0.2em 1em 0.2em; text-align: left; }\n"
              << "tr.line td { border-top: 1px solid black; }\n"
              << "</style>\n\n"
              << "</head>\n\n"
              << "<body>\n\n";

    if (m_basicExportPage->exportTitle()) {
        outStream << "<h1>" << targetFileBaseName().toHtmlEscaped() << "</h1>\n\n";
    }

    if (m_exportDatasetsPage->exportGenericData()) {
        // Add tournament meta data

        outStream << "<section>\n\n";

        if (m_exportFormattingPage->exportHeadlines()) {
            outStream << "<h2>" << tr("Allgemeine Daten") << "</h2>\n\n";
        }
        outStream << "<p>" << tr("Protokoll erstellt am %1 um %2 Uhr.").arg(
                                 QDateTime::currentDateTime().toString(m_dateFormat),
                                 QDateTime::currentDateTime().toString(m_timeFormat)) << "</p>\n\n";

        if (m_exportFormattingPage->exportSubHeadlines()) {
            outStream << "<h3>Muckturnier</h3>\n\n";
        }

        outStream << "<table>\n";
        outStream << "<tr><td>" << tr("Homepage:") << "</td>"
                  << "<td><a href=\"" << Urls::homepage << "\">" << Urls::homepage << "</a></td>"
                  << "</tr>\n";
        if (QStringLiteral(MT_VERSION) == m_tournamentSettings->muckturnierVersion()) {
            outStream << "<tr>"
                      << "<td>" << tr("Turnierdatenbank und Protokoll erstellt mit:") << "</td>"
                      << "<td>Version " << MT_VERSION << "</td>"
                      << "</tr>\n";
        } else {
            outStream << "<tr><td>" << tr("Turnierdatenbank erstellt mit:")
                      << "</td><td>" << tr("Version ")
                      << m_tournamentSettings->muckturnierVersion() << "</td></tr>\n"
                      << "<tr><td>" << tr("Protokoll erstellt mit:")
                      << "</td><td>" << tr("Version ")
                      << MT_VERSION << "</td></tr>\n";
        }
        outStream << "</table>\n\n";

        if (m_exportFormattingPage->exportSubHeadlines()) {
            outStream << "<h3>" << tr("Turnier") << "</h3>\n\n";
        }

        outStream << "<table>\n"
                  << "<tr><td>" << tr("Turniertyp:") << "</td>"
                  << "<td>" << (m_tournamentSettings->isPairMode() ? tr("Feste Paare")
                                                                   : tr("Einzelne Spieler"))
                  << "</td></tr>\n"
                  << "<tr><td>" << tr("Punkte pro Bobbl:") << "</td>"
                  << "<td>" << QString::number(m_tournamentSettings->boogerScore())
                  << "</td></tr>\n"
                  << "<tr><td>" << tr("Bobbl pro Runde:") << "</td>"
                  << "<td>" << QString::number(m_tournamentSettings->boogersPerRound())
                  << "</td></tr>\n"
                  << "<tr class=\"line\"><td>"
                  << (m_tournamentSettings->isPairMode() ? tr("Teilnehmende Paare:")
                                                         : tr("Teilnehmende Spieler:"))
                  << "</td>"
                  << "<td>" << QString::number(m_db->pairsOrPlayersCount())
                  << "</td></tr>\n"
                  << "<tr><td>" << tr("Gespielte Runden:") << "</td>"
                  << "<td>" << QString::number(m_db->allRounds()) << "</td></tr>\n"
                  << "</table>\n\n";

        outStream << "</section>\n\n";
    }

    if (m_exportDatasetsPage->exportRanking()) {
        // Add the ranking

        outStream << "<section>\n\n";

        if (m_exportFormattingPage->exportHeadlines()) {
            outStream << "<h2>" << tr("Rangliste nach %n Runde(n)", "", m_db->allRounds())
                      << "</h2>\n\n";
        }
        outStream << "<table>\n"
                  << "<thead>\n"
                  << "<tr>"
                  << "<th>" << tr("Platz") << "</th>"
                  << "<th>" << (m_tournamentSettings->isPairMode() ? tr("Paar")
                                                                   : tr("Spieler")) << "</th>"
                  << "<th>" << tr("Bobbl") << "</th>"
                  << "<th>" << tr("Tore") << "</th>";
        if (m_tournamentSettings->includeOpponentGoals()) {
            outStream << "<th>" << tr("Gegn. Tore") << "</th>";
        }
        outStream << "</tr>\n</thead>\n"
                  << "<tbody>\n";

        int lastRank = -1;
        QString style;

        const auto [ ranking, success ] = m_db->ranking(-1);
        if (! success) {
            return;
        }

        const auto goalsData = prepareGoals(ranking);

        for (int i = 0; i < ranking.count(); i++) {
            const auto &currentRank = ranking.at(i);
            const auto &goals = goalsData.at(i);

            if (currentRank.rank != lastRank) {
                style = QStringLiteral(" class=\"line\"");
                lastRank = currentRank.rank;
            } else {
                style.clear();
            }

            outStream << "<tr" << style << "><td>" << QString::number(currentRank.rank) << "</td>";
            outStream << "<td>" << currentRank.name << "</td>";
            outStream << "<td>" << currentRank.boogers << "</td>";
            outStream << "<td>" << goals.goals << "</td>";
            if (m_tournamentSettings->includeOpponentGoals()) {
                outStream << "<td>" << goals.opponentGoals << "</td>";
            }
            outStream << "</tr>\n";
        }
        outStream << "</tbody>\n"
                  << "</table>\n\n";

        outStream << "</section>\n\n";
    }

    if (m_exportDatasetsPage->exportRoundResults()) {
        outStream << "<section>\n\n";

        if (m_exportFormattingPage->exportHeadlines()) {
            outStream << "<h2>" << tr("Einzelergebnisse") << "</h2>\n\n";
        }

        switch (m_tournamentSettings->scoreType()) {
        case Tournament::ScoreType::Horizontal:
            saveHtmlHorizontalRoundResults(outStream);
            break;
        case Tournament::ScoreType::Vertical:
            saveHtmlVerticalRoundResults(outStream);
            break;
        }

        outStream << "</section>\n\n";
    }

    // Close HTML
    outStream << "</body>\n\n"
              << "</html>\n";
}

void ExportTournamentDialog::saveHtmlHorizontalRoundResults(QTextStream &outStream)
{
    const int boogersPerRound = m_tournamentSettings->boogersPerRound();

    const auto selectedRounds = m_exportDatasetsPage->roundSelection();
    for (const int round : selectedRounds) {
        if (m_exportFormattingPage->exportSubHeadlines()) {
            outStream << "<h3>" << tr("Spielergebnisse der %1. Runde").arg(round) << "</h3>\n\n";
        }
        outStream << "<table>\n<thead>\n<tr>";
        if (! m_tournamentSettings->noTableNumbers()) {
            outStream << "<th>" << tr("Tisch") << "</th>";
        }
        outStream << "<th>" << tr("Paar") << "</th>";

        for (int booger = 1; booger <= boogersPerRound; booger++) {
            outStream << "<th>" << tr("%1. Bobbl").arg(booger) << "</th>";
        }
        outStream << "</tr>\n</thead>\n<tbody>\n";

        const auto [ score, success ] = m_db->scores(round);
        if (! success) {
            QDialog::reject();
            return;
        }

        for (int i = 0; i < score.scores.count(); i++) {
            // Table numbers / top lines
            if (i % 2 == 0) {
                outStream << "<tr class=\"line\">";
                if (! m_tournamentSettings->noTableNumbers()) {
                    outStream << "<td rowspan=\"2\">" << score.tables.at(i / 2) << "</td>";
                }
            } else {
                outStream << "<tr>";
            }

            // Pair names
            outStream << "<td>" << score.pairs.at(i).toHtmlEscaped() << "</td>";

            // Results
            const auto &pairResults = score.scores.at(i);
            for (const int result : pairResults) {
                outStream << "<td>" << displayScore(result) << "</td>";
            }

            outStream << "</tr>\n";
        }

        outStream << "</tbody>\n" << "</table>\n\n";
    }
}

void ExportTournamentDialog::saveHtmlVerticalRoundResults(QTextStream &outStream)
{
    const int boogersPerRound = m_tournamentSettings->boogersPerRound();

    const auto selectedRounds = m_exportDatasetsPage->roundSelection();
    for (const int round : selectedRounds) {
        if (m_exportFormattingPage->exportSubHeadlines()) {
            outStream << "<h3>" << tr("Spielergebnisse der %1. Runde").arg(round) << "</h3>\n\n";
        }
        outStream << "<table>\n<thead>\n<tr>";
        if (! m_tournamentSettings->noTableNumbers()) {
            outStream << "<th>" << tr("Tisch") << "</th>";
        }
        outStream << "<th>" << tr("Bobbl") << "</th><th>" << tr("Paar 1") << "</th><th>"
                  << tr("Paar 2") << "</th></tr>\n</thead>\n<tbody>\n";

        const auto [ score, success ] = m_db->scores(round);
        if (! success) {
            QDialog::reject();
            return;
        }

        for (int i = 0; i < score.scores.count(); i += 2) {
            outStream << "<tr class=\"line\">";

            // Table numbers
            if (! m_tournamentSettings->noTableNumbers()) {
                outStream << "<td rowspan=\"" << (boogersPerRound + 1) << "\">"
                            << score.tables.at(i / 2) << "</td>";
            }

            // Pair names
            outStream << "<td></td>"
                        << "<td>" << score.pairs.at(i).toHtmlEscaped() << "</td>"
                        << "<td>" << score.pairs.at(i + 1).toHtmlEscaped() << "</td></tr>";

            for (int booger = 1; booger <= boogersPerRound; booger++) {
                // Booger label
                outStream << "<tr><td>" << booger << ".</td><td>";

                // Pair results
                outStream << displayScore(score.scores.at(i).at(booger - 1)) << "</td><td>"
                            << displayScore(score.scores.at(i + 1).at(booger - 1)) << "</td>"
                            << "</tr>\n";
            }
        }

        outStream << "</tbody>\n" << "</table>\n\n";
    }
}

void ExportTournamentDialog::saveCsvFile(QTextStream &outStream)
{
    if (m_basicExportPage->exportTitle()) {
        outStream << m_basicExportPage->title() << "\n\n";
    }

    if (m_exportDatasetsPage->exportGenericData()) {
        // Add tournament meta data

        if (m_exportFormattingPage->exportHeadlines()) {
            outStream << tr("Allgemeine Daten") << "\n\n";
        }

        outStream << tr("Protokoll erstellt am %1 um %2 Uhr.").arg(
                        QDateTime::currentDateTime().toString(m_dateFormat),
                        QDateTime::currentDateTime().toString(m_timeFormat))
                  << "\n\n";

        if (m_exportFormattingPage->exportSubHeadlines()) {
            outStream << "Muckturnier\n\n";
        }

        outStream << tr("Homepage:") << "\t" << Urls::homepage << "\n";
        if (QStringLiteral(MT_VERSION) == m_tournamentSettings->muckturnierVersion()) {
            outStream << tr("Turnierdatenbank und Protokoll erstellt mit: \tVersion ")
                      << MT_VERSION << "\n";
        } else {
            outStream << tr("Turnierdatenbank erstellt mit:\tVersion ")
                      << m_tournamentSettings->muckturnierVersion() << "\n"
                      << tr("Protokoll erstellt mit:\tVersion ")
                      << MT_VERSION << "\n";
        }
        outStream << "\n";

        if (m_exportFormattingPage->exportSubHeadlines()) {
            outStream << tr("Turnier") << "\n\n";
        }

        outStream << tr("Turniertyp:")
                  << "\t" << (m_tournamentSettings->isPairMode() ? tr("Feste Paare")
                                                                 : tr("Einzelne Spieler"))
                  << "\n"
                  << tr("Punkte pro Bobbl:")
                  << "\t" << QString::number(m_tournamentSettings->boogerScore()) << "\n"
                  << tr("Bobbl pro Runde:")
                  << "\t" << QString::number(m_tournamentSettings->boogersPerRound()) << "\n"
                  << (m_tournamentSettings->isPairMode() ? tr("Teilnehmende Paare:")
                                                         : tr("Teilnehmende Spieler:"))
                  << "\t" << QString::number(m_db->pairsOrPlayersCount()) << "\n"
                  << tr("Gespielte Runden:") << "\t"
                  << QString::number(m_db->allRounds()) << "\n\n";
    }

    if (m_exportDatasetsPage->exportRanking()) {
        // Add the ranking

        if (m_exportFormattingPage->exportHeadlines()) {
            outStream << tr("Rangliste nach %n Runde(n)", "", m_db->allRounds()) << "\n\n";
        }
        outStream << tr("Platz") << "\t"
                  << (m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler")) << "\t"
                  << tr("Bobbl") << "\t"
                  << tr("Tore");
        if (m_tournamentSettings->includeOpponentGoals()) {
            outStream << "\t" << tr("Gegn. Tore");
        }
        outStream << "\n";

        const auto [ ranking, success ] = m_db->ranking(-1);
        if (! success) {
            return;
        }

        const auto goalsData = prepareGoals(ranking);

        for (int i = 0; i < ranking.count(); i++) {
            const auto &currentRank = ranking.at(i);
            const auto &goals = goalsData.at(i);

            outStream << QString::number(currentRank.rank);
            outStream << "\t" << currentRank.name;
            outStream << "\t" << QString::number(currentRank.boogers);
            outStream << "\t" << goals.goals;
            if (m_tournamentSettings->includeOpponentGoals()) {
                outStream << "\t" << goals.opponentGoals;
            }
            outStream << "\n";
        }
        outStream << "\n";
    }

    if (m_exportDatasetsPage->exportRoundResults()) {
        if (m_exportFormattingPage->exportHeadlines()) {
            outStream << tr("Einzelergebnisse") << "\n\n";
        }
        switch (m_tournamentSettings->scoreType()) {
        case Tournament::ScoreType::Horizontal:
            saveCsvHorizontalRoundResults(outStream);
            break;
        case Tournament::ScoreType::Vertical:
            saveCsvVerticalRoundResults(outStream);
            break;
        }
    }
}

void ExportTournamentDialog::saveCsvHorizontalRoundResults(QTextStream &outStream)
{
    const QString tableTab = m_tournamentSettings->noTableNumbers() ? QString()
                                                                    : QLatin1String("\t");

    const auto selectedRounds = m_exportDatasetsPage->roundSelection();
    for (const int round : selectedRounds) {
        if (m_exportFormattingPage->exportSubHeadlines()) {
            outStream << tr("Spielergebnisse der %1. Runde").arg(round) << "\n";
        }
        if (! m_tournamentSettings->noTableNumbers()) {
            outStream << tr("Tisch\t");
        }
        outStream << tr("Paar");
        for (int booger = 1; booger <= m_tournamentSettings->boogersPerRound(); booger++) {
            outStream << tr("\t%1. Bobbl").arg(booger);
        }
        outStream << "\n";

        const auto [ score, success ] = m_db->scores(round);
        if (! success) {
            QDialog::reject();
            return;
        }

        for (int i = 0; i < score.scores.count(); i += 2) {
            // Table number
            if (! m_tournamentSettings->noTableNumbers()) {
                outStream << score.tables.at(i / 2) << "\t";
            }

            for (int pair = 1; pair <= 2; pair++) {
                // Pair names
                if (pair == 2) {
                    outStream << tableTab;
                }
                outStream << score.pairs.at(i + pair - 1);

                // Results
                for (int result : score.scores.at(i + pair - 1)) {
                    outStream << "\t" << displayScore(result);
                }

                outStream << "\n";
            }
        }

        outStream << "\n";
    }
}

void ExportTournamentDialog::saveCsvVerticalRoundResults(QTextStream &outStream)
{
    const int boogersPerRound = m_tournamentSettings->boogersPerRound();
    const QString tableTab = m_tournamentSettings->noTableNumbers() ? QString()
                                                                    : QLatin1String("\t");

    const auto selectedRounds = m_exportDatasetsPage->roundSelection();
    for (const int round : selectedRounds) {
        if (m_exportFormattingPage->exportSubHeadlines()) {
            outStream << tr("Spielergebnisse der %1. Runde").arg(round) << "\n";
        }
        if (! m_tournamentSettings->noTableNumbers()) {
            outStream << tr("Tisch") << "\t";
        }
        outStream << tr("Bobbl") << "\t" << tr("Paar 1") << "\t"<< tr("Paar 2") << "\n";

        const auto [ score, success ] = m_db->scores(round);
        if (! success) {
            QDialog::reject();
            return;
        }

        for (int i = 0; i < score.scores.count(); i += 2) {
            // Table numbers
            if (! m_tournamentSettings->noTableNumbers()) {
                outStream << score.tables.at(i / 2);
            }

            // Pair names
            outStream << "\t" << tableTab << score.pairs.at(i).toHtmlEscaped() << "\t"
                                          << score.pairs.at(i + 1).toHtmlEscaped() << "\n";

            for (int booger = 1; booger <= boogersPerRound; booger++) {
                // Booger label
                outStream << tableTab << booger << ".\t";

                // Pair results
                outStream << displayScore(score.scores.at(i).at(booger - 1)) << "\t"
                          << displayScore(score.scores.at(i + 1).at(booger - 1)) << "\n";
            }
        }

        outStream << "\n";
    }
}

QString ExportTournamentDialog::displayScore(int score) const
{
    return m_exportFormattingPage->displayBoogers() ? m_stringTools->boogify(score)
                                                    : QString::number(score);
}

QVector<ExportTournamentDialog::GoalsData> ExportTournamentDialog::prepareGoals(
    const QVector<Database::RankData> &ranking) const
{
    QVector<GoalsData> data;
    data.reserve(ranking.count());

    const auto unimportantGoalsHandling = m_exportFormattingPage->unimportantGoalsHandling();
    QVector<bool> goalsImportant;
    QVector<bool> opponentGoalsImportant;
    switch (unimportantGoalsHandling) {
    case ExportFormattingPage::UnimportantGoalsHandling::Display:
        break;
    case ExportFormattingPage::UnimportantGoalsHandling::Brace:
    case ExportFormattingPage::UnimportantGoalsHandling::Hide:
        const auto [ a, b ] = m_db->goalsImportance(ranking);
        goalsImportant = a;
        opponentGoalsImportant = b;
    }

    for (int i = 0; i < ranking.count(); i++) {
        GoalsData entry;
        const auto &currentRank = ranking.at(i);

        switch (unimportantGoalsHandling) {
        case ExportFormattingPage::UnimportantGoalsHandling::Display:
            entry.goals = QString::number(currentRank.goals);
            entry.opponentGoals = QString::number(currentRank.opponentGoals);
            break;
        case ExportFormattingPage::UnimportantGoalsHandling::Brace:
            entry.goals = goalsImportant.at(i)
                ? QString::number(currentRank.goals)
                : QStringLiteral("(%1)").arg(currentRank.goals);
            entry.opponentGoals = opponentGoalsImportant.at(i)
                ? QString::number(currentRank.opponentGoals)
                : QStringLiteral("(%1)").arg(currentRank.opponentGoals);
            break;
        case ExportFormattingPage::UnimportantGoalsHandling::Hide:
            if (goalsImportant.at(i)) {
                entry.goals = QString::number(currentRank.goals);
            }
            if (opponentGoalsImportant.at(i)) {
                entry.opponentGoals = QString::number(currentRank.opponentGoals);
            }
            break;
        }

        data.append(entry);
    }

    return data;
}
