/* SPDX-FileCopyrightText: 2010-2021 Tobias Leupold <tobias.leupold@gmx.de>

   SPDX-License-Identifier: GPL-3.0-or-later
*/

// Local includes

#include "RankingPage.h"

#include "shared/SharedObjects.h"
#include "shared/SearchEngine.h"
#include "shared/SearchWidget.h"
#include "shared/TableDelegate.h"
#include "shared/Settings.h"
#include "shared/TournamentSettings.h"

// Qt includes
#include <QDebug>
#include <QGroupBox>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QTableWidget>
#include <QHeaderView>
#include <QMenu>
#include <QPushButton>
#include <QComboBox>

RankingPage::RankingPage(QWidget *parent, SharedObjects *sharedObjects)
    : QWidget(parent),
      m_db(sharedObjects->database()),
      m_searchEngine(sharedObjects->searchEngine()),
      m_tournamentSettings(sharedObjects->tournamentSettings())
{
    QVBoxLayout *layout = new QVBoxLayout(this);

    m_noResultsLabel = new QLabel(tr("<i>Es liegen noch keine Ergebnisse vor</i>"));
    m_noResultsLabel->setAlignment(Qt::AlignCenter);
    m_noResultsLabel->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Expanding);
    layout->addWidget(m_noResultsLabel);

    m_settingsBox = new QGroupBox;
    m_settingsBox->setProperty("notitle", true);
    QHBoxLayout *settingsBoxLayout = new QHBoxLayout(m_settingsBox);
    m_settingsBox->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Maximum);
    layout->addWidget(m_settingsBox);
    m_settingsBox->hide();

    m_upToRoundLabel = new QLabel(tr("Rangliste für Runde"));
    m_upToRoundLabel->setAlignment(Qt::AlignRight);
    m_upToRoundLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    settingsBoxLayout->addWidget(m_upToRoundLabel);
    m_upToRoundLabel->hide();

    m_upToRound = new QComboBox;
    m_upToRound->setSizeAdjustPolicy(QComboBox::AdjustToContents);
    m_upToRound->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    connect(m_upToRound, QOverload<int>::of(&QComboBox::currentIndexChanged),
            this, [this](int round) { reload(round + 1); });
    settingsBoxLayout->addWidget(m_upToRound);
    m_upToRound->hide();

    settingsBoxLayout->addStretch();

    QPushButton *settingsButton = new QPushButton(tr("Einstellungen"));
    QMenu *settingsMenu = new QMenu(this);

    m_includeOpponentGoals = settingsMenu->addAction(tr("Gegnerische Tore berücksichtigen"));
    m_includeOpponentGoals->setCheckable(true);
    connect(m_includeOpponentGoals, &QAction::toggled,
            this, &RankingPage::includeOpponentGoalsChanged);

    settingsMenu->addSeparator();

    m_grayOutUnimportantGoals = settingsMenu->addAction(tr("Nicht entscheidende Tore ausgrauen"));
    m_grayOutUnimportantGoals->setCheckable(true);
    connect(m_grayOutUnimportantGoals, &QAction::toggled,
            this, &RankingPage::grayOutUnimportantGoalsChanged);

    settingsButton->setMenu(settingsMenu);
    settingsBoxLayout->addWidget(settingsButton);

    m_rankingBox = new QGroupBox;
    QVBoxLayout *rankingBoxLayout = new QVBoxLayout(m_rankingBox);
    layout->addWidget(m_rankingBox);
    m_rankingBox->hide();

    m_isSearchLabel = new QLabel;
    m_isSearchLabel->setAlignment(Qt::AlignCenter);
    m_isSearchLabel->hide();
    rankingBoxLayout->addWidget(m_isSearchLabel);

    m_interimLabel = new QLabel;
    m_interimLabel->setWordWrap(true);
    rankingBoxLayout->addWidget(m_interimLabel);

    m_ranking = new QTableWidget;
    m_ranking->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_ranking->setItemDelegate(new TableDelegate(this, TableDelegate::HighlightMode::Custom));
    m_ranking->verticalHeader()->hide();
    m_ranking->setShowGrid(false);
    m_ranking->setWordWrap(false);
    m_ranking->setSelectionMode(QTableWidget::NoSelection);
    m_ranking->setFocusPolicy(Qt::NoFocus);
    m_ranking->setColumnCount(5);
    rankingBoxLayout->addWidget(m_ranking);

    m_multipleRanks = new QLabel;
    m_multipleRanks->setWordWrap(true);
    m_multipleRanks->hide();
    rankingBoxLayout->addWidget(m_multipleRanks);

    m_searchWidget = new SearchWidget(QStringLiteral("ranking_page_search"), sharedObjects);
    connect(m_searchWidget, &SearchWidget::searchChanged, this, &RankingPage::updateSearch);
    rankingBoxLayout->addWidget(m_searchWidget);
}

void RankingPage::setupGui()
{
    m_includeOpponentGoals->blockSignals(true);
    m_grayOutUnimportantGoals->blockSignals(true);

    m_includeOpponentGoals->setChecked(m_tournamentSettings->includeOpponentGoals());
    m_includeOpponentGoals->setEnabled(m_db->isChangeable());

    m_grayOutUnimportantGoals->setChecked(m_tournamentSettings->grayOutUnimportantGoals());

    m_includeOpponentGoals->blockSignals(false);
    m_grayOutUnimportantGoals->blockSignals(false);

    m_ranking->setHorizontalHeaderLabels( { tr("Platz"),
                                            m_tournamentSettings->isPairMode() ? tr("Paar")
                                                                               : tr("Spieler"),
                                            tr("Bobbl"),
                                            tr("Tore"),
                                            tr("Gegn. Tore") } );
    m_ranking->horizontalHeaderItem(2)->setWhatsThis(tr(
        "In dieser Spalte werden die jeweils gewonnenen Bobbl angezeigt – je mehr, desto besser"));

    m_ranking->horizontalHeaderItem(3)->setWhatsThis(tr(
        "„Geschossene Tore“ sind die durch %1 in den verlorenen Bobbln erreichten Punkte – je "
        "mehr, desto besser.").arg(
        m_tournamentSettings->isPairMode() ? tr("das jeweilige Paar")
                                           : tr("den jeweiligen Spieler")));

    m_ranking->horizontalHeaderItem(4)->setWhatsThis(tr(
        "„Gegenerische geschossene Tore“ sind die Punkte, die die Gegener in den gewonnenen "
        "Bobbln des jeweiligen %1 erreicht haben – je weniger, desto besser.").arg(
        m_tournamentSettings->isPairMode() ? tr("Paars") : tr("Spielers")));

    m_ranking->setColumnHidden(4, ! m_includeOpponentGoals->isChecked());
    m_ranking->resizeColumnsToContents();

    reload();
}

void RankingPage::dataChanged()
{
    m_upToRound->blockSignals(true);
    m_upToRound->clear();
    for (int i = 1; i <= m_db->allRounds(); i++) {
        m_upToRound->addItem(QString::number(i));
    }
    m_upToRound->setCurrentIndex(m_db->allRounds() - 1);
    m_upToRound->blockSignals(false);

    if (m_upToRound->count() == 1) {
        m_upToRound->hide();
        m_upToRoundLabel->hide();
    } else {
        m_upToRound->show();
        m_upToRoundLabel->show();
    }

    reload(m_db->allRounds());
}

void RankingPage::includeOpponentGoalsChanged()
{
    const bool state = m_includeOpponentGoals->isChecked();
    m_tournamentSettings->setIncludeOpponentGoals(state);
    m_ranking->setColumnHidden(4, ! state);
    reload(m_upToRound->currentIndex() + 1);
}

void RankingPage::reload(int upToRound)
{
    const bool localTournamentStarted = m_db->localTournamentStarted();
    m_noResultsLabel->setVisible(! localTournamentStarted);
    m_settingsBox->setVisible(localTournamentStarted);
    m_rankingBox->setVisible(localTournamentStarted);
    if (! localTournamentStarted) {
        return;
    }

    m_searchWidget->saveSearch();

    QVector<int> incompleteRounds;
    const auto roundsState = m_db->roundsState();
    const auto roundsStateKeys = roundsState.keys();
    for (const int round : roundsStateKeys) {
        if (! roundsState.value(round)) {
            incompleteRounds.append(round);
        }
    }
    const int missingRounds = incompleteRounds.count();

    QString interimText;
    if (missingRounds > 0) {
        emit enableExport(false);
        interimText = tr("<table><tr><td><b>Achtung!</b></td><td>Das ist nur ein Zwischenergebnis. "
                         "Es fehlen noch Ergebnisse aus ");
        if (missingRounds == 1) {
            interimText += tr("Runde %1.").arg(QString::number(incompleteRounds.at(0)));
        } else if (missingRounds == 2) {
            interimText += tr("den Runden %1 und %2.").arg(QString::number(incompleteRounds.at(0)),
                                                           QString::number(incompleteRounds.at(1)));
        } else {
            interimText += tr("den Runden ");
            for (int i = 0; i < missingRounds - 1; i++) {
                interimText += tr("%1, ").arg(QString::number(incompleteRounds.at(i)));
            }
            interimText = interimText.left(interimText.size() - 2);
            interimText += tr(" und %1.").arg(QString::number(incompleteRounds.last()));
        }
        interimText += QStringLiteral(
            "<br/>%1 <span style=\"color:red\">%2</span></td></tr></table>").arg(
            m_tournamentSettings->isPairMode()
                ? tr("Paare, von denen noch Spielstände fehlen, sind")
                : tr("Spieler, von denen noch Spielstände fehlen, sind"),
            tr("rot markiert"));
    } else {
        emit enableExport(true);
    }

    if (upToRound < m_db->allRounds()) {
        if (m_db->roundsState()[upToRound]) {
            m_rankingBox->setTitle(tr("Zwischenstand: Rangliste nach Runde %1").arg(upToRound));
        } else {
            m_rankingBox->setTitle(tr("Zwischenstand: Rangliste im Verlauf von Runde %1").arg(
                                      upToRound));
        }
        interimText = tr("<table><tr><td><b>Achtung!</b></td><td>Das ist nicht die aktuelle "
                         "Rangliste! Um diese anzuzeigen <b>bitte Runde %1 auswählen!</b></td>"
                         "</tr></table>").arg(m_db->allRounds())
                      + interimText;
    } else {
        if (! m_db->roundsState()[m_db->allRounds()]) {
            m_rankingBox->setTitle(tr("Aktuelle Rangliste im Verlauf der %1. Runde").arg(
                                      QString::number(m_db->allRounds())));
        } else {
            if (m_db->allRounds() == 1) {
                if (m_db->isChangeable()) {
                    m_rankingBox->setTitle(tr("Aktuelle Rangliste nach einer Runde"));
                } else {
                    m_rankingBox->setTitle(tr("Rangliste nach einer Runde"));
                }
            } else {
                if (m_db->isChangeable()) {
                    m_rankingBox->setTitle(tr("Aktuelle Rangliste nach %1 Runden").arg(
                                           QString::number(m_db->allRounds())));
                } else {
                    m_rankingBox->setTitle(tr("Rangliste nach %1 Runden").arg(
                                           QString::number(m_db->allRounds())));
                }
            }
        }
    }

    if (interimText.isEmpty()) {
        m_interimLabel->hide();
    } else {
        m_interimLabel->setText(interimText);
        m_interimLabel->show();
    }

    const auto [ ranking, success ] = m_db->ranking(upToRound);
    if (! success) {
        return;
    }

    const int allBoogers = upToRound * m_tournamentSettings->boogersPerRound();

    m_ranking->setRowCount(ranking.count());

    int row = 0;
    int lastRank = 0;
    int color = 1;
    bool drawTopLine = false;
    QMap<int, int> multipleRanks;

    for (const Database::RankData &currentRank : ranking) {
        if (currentRank.rank != lastRank) {
            color++;
            color = color % 2;
            if (lastRank != 0) {
                drawTopLine = true;
            }
        } else {
            // This will insert rank keys if they don't exist yet and increment them if they exist
            multipleRanks[currentRank.rank]++;
        }
        lastRank = currentRank.rank;

        QTableWidgetItem *rankCell = new QTableWidgetItem(QString::number(currentRank.rank));
        QTableWidgetItem *nameCell = new QTableWidgetItem(currentRank.name);
        QTableWidgetItem *boogersCell = new QTableWidgetItem(QString::number(currentRank.boogers));
        QTableWidgetItem *goalsCell = new QTableWidgetItem(QString::number(currentRank.goals));
        QTableWidgetItem *opponentGoalsCell = new QTableWidgetItem(
                                                      QString::number(currentRank.opponentGoals));

        if (currentRank.playedBoogers != allBoogers) {
            rankCell->setForeground(Qt::red);
            nameCell->setForeground(Qt::red);
            boogersCell->setForeground(Qt::red);
            goalsCell->setForeground(Qt::red);
            opponentGoalsCell->setForeground(Qt::red);
        }

        if (color) {
            rankCell->setData(TableDelegate::Display::DrawBackground, QVariant(true));
            nameCell->setData(TableDelegate::Display::DrawBackground, QVariant(true));
            boogersCell->setData(TableDelegate::Display::DrawBackground, QVariant(true));
            goalsCell->setData(TableDelegate::Display::DrawBackground, QVariant(true));
            opponentGoalsCell->setData(TableDelegate::Display::DrawBackground, QVariant(true));
        }

        if (drawTopLine && row > 0) {
            rankCell->setData(TableDelegate::Display::DrawTopLine, QVariant(true));
            nameCell->setData(TableDelegate::Display::DrawTopLine, QVariant(true));
            boogersCell->setData(TableDelegate::Display::DrawTopLine, QVariant(true));
            goalsCell->setData(TableDelegate::Display::DrawTopLine, QVariant(true));
            opponentGoalsCell->setData(TableDelegate::Display::DrawTopLine, QVariant(true));
            drawTopLine = false;
        }

        m_ranking->setItem(row, 0, rankCell);
        m_ranking->setItem(row, 1, nameCell);
        m_ranking->setItem(row, 2, boogersCell);
        m_ranking->setItem(row, 3, goalsCell);
        m_ranking->setItem(row, 4, opponentGoalsCell);

        row++;
    }

    if (m_grayOutUnimportantGoals->isChecked()) {
        grayOutUnimportantGoals(ranking);
    }

    if (multipleRanks.count() > 0) {
        QString multipleRanksText = tr("Mehrfach besetzte Ranglistenplätze: ");
        const auto multipleRanksKeys = multipleRanks.keys();
        for (const int rank : multipleRanksKeys) {
            multipleRanksText += tr("<b>%1</b> (%2x), ").arg(
                                    QString::number(rank),
                                    QString::number(multipleRanks[rank] + 1));
        }
        multipleRanksText.chop(2);
        m_multipleRanks->setText(multipleRanksText);
        m_multipleRanks->show();
    } else {
        m_multipleRanks->hide();
    }

    m_ranking->resizeColumnsToContents();
    m_searchWidget->restoreSearch();
}

void RankingPage::grayOutUnimportantGoals(const QVector<Database::RankData> &ranking)
{
    const auto [ goalsImportant, opponentGoalsImportant ] = m_db->goalsImportance(ranking);

    for (int i = 0; i < goalsImportant.count(); i++) {
        const auto goalsUnimportant = ! goalsImportant.at(i);
        const auto opponentGoalsUnimportant = ! opponentGoalsImportant.at(i);

        if (goalsUnimportant || opponentGoalsUnimportant) {
            const QColor color = grayedOutColor(m_ranking->item(i, 0)->foreground().color());

            if (goalsUnimportant) {
                m_ranking->item(i, 3)->setForeground(color);
            }

            if (opponentGoalsUnimportant) {
                m_ranking->item(i, 4)->setForeground(color);
            }
        }
    }
}

QColor RankingPage::grayedOutColor(const QColor &color)
{
    // We need the color's int representation because a QColor can't be a QHash key
    const QRgb rgb = color.rgb();

    // Calculate the requested color grayed out if we don't have it yet
    if (! m_grayedOutColors.contains(rgb)) {
        QColor processedColor = color.toCmyk();
        processedColor.setCmyk((int) ((float) processedColor.cyan() / 2.8),
                               (int) ((float) processedColor.magenta() / 2.8),
                               (int) ((float) processedColor.yellow() / 2.8),
                               (int) ((float) processedColor.black() / 2.8));
        m_grayedOutColors.insert(rgb, processedColor);
    }

    // Return the grayed out color
    return m_grayedOutColors.value(rgb);
}

void RankingPage::updateSearch(const QString &searchText)
{
    if (searchText.isEmpty()) {
        for (int row = 0; row < m_ranking->rowCount(); row++) {
            m_ranking->showRow(row);
        }
        m_ranking->resizeColumnsToContents();
        m_isSearchLabel->hide();
        return;
    }

    int matches = 0;
    const bool phoneticSearch = m_searchWidget->phoneticSearch();
    m_searchEngine->setSearchTerm(searchText);
    for (int row = 0; row < m_ranking->rowCount(); row++) {
        const bool rowMatches = m_searchEngine->checkMatch(m_ranking->item(row, 1)->text(),
                                                           SearchEngine::SearchType::Default,
                                                           phoneticSearch);
        m_ranking->setRowHidden(row, ! rowMatches);
        matches += rowMatches;
    }
    m_searchWidget->updateMatches(matches);
    m_ranking->resizeColumnsToContents();

    m_isSearchLabel->setText(tr("(Suchergebnis für „%1“)").arg(searchText.simplified()));
    m_isSearchLabel->show();
}

void RankingPage::resetSearch()
{
    m_searchWidget->clear();
}

void RankingPage::grayOutUnimportantGoalsChanged()
{
    m_tournamentSettings->setGrayOutUnimportantGoals(m_grayOutUnimportantGoals->isChecked());
    reload(m_upToRound->currentIndex() + 1);
}
