/* SPDX-FileCopyrightText: 2018-2020 Tobias Leupold <tobias.leupold@gmx.de>

   SPDX-License-Identifier: GPL-3.0-or-later
*/

// Local includes

#include "PlayersComboBox.h"
#include "PlayersLineEdit.h"
#include "PlayersListView.h"

#include "shared/SearchEngine.h"

// Qt includes
#include <QDebug>
#include <QStringListModel>
#include <QCompleter>

PlayersComboBox::PlayersComboBox(SearchEngine *searchEngine, int index, QWidget *parent)
    : NumbersComboBox(parent),
      m_searchEngine(searchEngine),
      m_index(index)
{
    setEditable(true);
    setInsertPolicy(QComboBox::NoInsert);
    setCompleter(nullptr);

    connect(this, QOverload<int>::of(&QComboBox::activated), this, &PlayersComboBox::finishSearch);
    connect(this, QOverload<int>::of(&QComboBox::currentIndexChanged),
            this, [this] { emit selectionChanged(m_index, currentNumber()); });

    m_matches = new QStringListModel(this);
    QCompleter *completer = new QCompleter(m_matches, this);
    completer->setCompletionMode(QCompleter::UnfilteredPopupCompletion);
    connect(completer, QOverload<const QString &>::of(&QCompleter::activated),
            this, &PlayersComboBox::completerActivated);

    PlayersListView *listView = new PlayersListView;
    completer->setPopup(listView);
    connect(listView, &PlayersListView::popupHidden, this, &PlayersComboBox::popupHidden);

    m_lineEdit = new PlayersLineEdit;
    m_lineEdit->setCompleter(completer);
    setLineEdit(m_lineEdit);

    disconnect(m_lineEdit, &QLineEdit::editingFinished, 0, 0);
    connect(m_lineEdit, &QLineEdit::textEdited, this, &PlayersComboBox::lineEditTextChanged);
    connect(m_lineEdit, &PlayersLineEdit::escapePressed,
            this, &PlayersComboBox::resetLineEditText);
    connect(m_lineEdit, &QLineEdit::editingFinished, this, &PlayersComboBox::processGivenText);

    m_isSearching = false;
}

void PlayersComboBox::lineEditTextChanged(const QString &text)
{
    if (! m_isSearching) {
        m_isSearching = true;
        m_completerActivated = false;
        m_lineEdit->displayInvalidInput(true);
        emit disableSaveScore();
    }

    QStringList matches;
    m_searchEngine->setSearchTerm(text);
    for(int i = 0; i < model()->rowCount(); i++) {
        const QString name = model()->index(i, 0).data().toString();
        if (m_searchEngine->checkMatch(name, SearchEngine::SearchType::Default, m_phoneticSearch)) {
            matches.append(name);
        }
    }
    m_matches->setStringList(matches);
}

void PlayersComboBox::resetLineEditText()
{
    m_lineEdit->setText(itemText(currentIndex()));
    finishSearch();
    m_lineEdit->setFocus();
}

void PlayersComboBox::processGivenText()
{
    if (! m_isSearching || m_completerActivated) {
        return;
    }

    if (m_matches->rowCount() == 1) {
        completerActivated(m_matches->index(0, 0).data().toString());
        return;
    }

    resetLineEditText();
}

void PlayersComboBox::completerActivated(const QString &text)
{
    m_completerActivated = true;
    setCurrentIndex(findText(text));
    finishSearch();
}

void PlayersComboBox::finishSearch()
{
    if (m_isSearching) {
        m_isSearching = false;
        m_lineEdit->displayInvalidInput(false);
    }
    emit checkSavePossible();
}

void PlayersComboBox::popupHidden()
{
    if (m_completerActivated || m_matches->rowCount() != 1) {
        return;
    }

    const QString onlyMatch = m_matches->index(0, 0).data().toString();
    if (m_lineEdit->text() != onlyMatch) {
        completerActivated(onlyMatch);
    }
}

void PlayersComboBox::setPhoneticSearch(bool state)
{
    m_phoneticSearch = state;
    if (m_isSearching) {
        resetLineEditText();
    }
}
