/* SPDX-FileCopyrightText: 2018-2020 Tobias Leupold <tobias.leupold@gmx.de>

   SPDX-License-Identifier: GPL-3.0-or-later
*/

#ifndef SERVER_H
#define SERVER_H

// Local includes
#include "shared/Markers.h"
#include "shared/Draw.h"

// Qt includes
#include <QTcpServer>

// Local classes
class SharedObjects;
class Database;
class TournamentSettings;
class RegistrationPage;
class MarkersWidget;
class Connection;
class Settings;

class Server : public QTcpServer
{
    Q_OBJECT

public:
    explicit Server(QObject *parent, int connectTimeout, SharedObjects *sharedObjects,
                    RegistrationPage *registrationPage, const QMap<int, QString> *scoreChecksums);
    void setChecksum(const QString *checksum);
    int clientCount() const;
    void close();

    // Changes broadcasting
    // ====================

    // Helper macro to keep the header a bit more tidy
    #define BROADCAST_FUNCTION(name, ...) void broadcast##name(__VA_ARGS__, int clientId = 0)
    #define BROADCAST_FUNCTION_NA(name) void broadcast##name(int clientId = 0)

    // Registration changes

    BROADCAST_FUNCTION(RegisterPlayers, const QString &name, int marker);
    BROADCAST_FUNCTION(DeletePlayers, const QString &name);
    BROADCAST_FUNCTION(RenamePlayers, const QString &name, const QString &newName,
                                      int markerId, const QString &markerName);
    BROADCAST_FUNCTION(AssemblePair, const QString &player1, const QString &player2,
                                     const QString &separator, int assignedMarker);
    BROADCAST_FUNCTION(UpdateMarker, const QString &name, const QString &markerName, int markerId);

    BROADCAST_FUNCTION(DeleteMarkedPlayers, const QString &markerName, int markerId);
    BROADCAST_FUNCTION_NA(DeleteUnmarkedPlayers);
    BROADCAST_FUNCTION_NA(DeleteAllPlayers);
    BROADCAST_FUNCTION(ListUpdateMarker, const QString &markerName, int markerId);
    BROADCAST_FUNCTION(ListRemoveMarker, const QString &markerName, int markerId);

    // Draw changes
    BROADCAST_FUNCTION(UpdateDraw, const QString &name, const Draw::Seat &seat, int marker);
    BROADCAST_FUNCTION(SetListDraw, const QVector<QString> &names, const QVector<Draw::Seat> &draw);
    BROADCAST_FUNCTION_NA(DeleteAllDraws);

    // Markers definition changes
    BROADCAST_FUNCTION(AddMarker, const QString &name, const QColor &color,
                                  Markers::MarkerSorting sorting);
    BROADCAST_FUNCTION(DeleteMarker, const QString &name, int id, int sequence);
    BROADCAST_FUNCTION(MoveMarker, const QString &name, int id, int sequence, int direction);
    BROADCAST_FUNCTION(EditMarker, const QString &name, int id, const QString &newName,
                                   const QColor &color, Markers::MarkerSorting sorting);

    // Special markers changes
    BROADCAST_FUNCTION(SetAbsentMarker, const QString &name, int id);
    BROADCAST_FUNCTION(SinglesManagementChange, const Markers::SinglesManagement &data);

    // Finishing the registration
    BROADCAST_FUNCTION_NA(FinishRegistration);

    // Ranking comparison
    // (Can't be requested by the server)

    // Score comparison
    // (Can't be requested by the server)

    #undef BROADCAST_FUNCTION
    #undef BROADCAST_FUNCTION_NA

public slots:
    void scoreChecksumChanged(int round);

signals:
    void tx();
    void rx();
    void protocolError(const QString &text);
    void message(const QString &text);
    void statusUpdate(const QString &text);
    void clientCountChanged(int count);
    void nameEdited();

protected:
    void incomingConnection(qintptr descriptor) override;

private slots:
    void checkHandshakeTimeout(Connection *connection);
    void removeConnection(Connection *connection);
    void processHandshake(Connection *connection, const QString &phase);
    void processClientRequest(Connection *connection, const QString &action,
                              const QJsonObject &jsonObject);

private: // Functions
    void broadcastMessage(QJsonObject &jsonObject, int clientId, const QString &messageText);
    void sendSingleReply(QJsonObject &jsonObject, Connection *connection,
                         const QString &messageText);

    QString clientString(Connection *connection) const;

    // Client request handlers
    // =======================

    // Helper macro to keep the header a bit more tidy
    #define REQUEST_HANDLER(name) \
            void process##name(const QJsonObject &jsonObject, const QString &client, int clientId)

    // Registration changes

    REQUEST_HANDLER(RegisterPlayers);
    REQUEST_HANDLER(RenamePlayers);
    REQUEST_HANDLER(AssemblePair);
    REQUEST_HANDLER(DeletePlayers);
    REQUEST_HANDLER(UpdatePlayersMarker);

    REQUEST_HANDLER(DeleteMarkedPlayers);
    REQUEST_HANDLER(DeleteUnmarkedPlayers);
    REQUEST_HANDLER(DeleteAllPlayers);
    REQUEST_HANDLER(UpdateListMarker);
    REQUEST_HANDLER(RemoveListMarker);

    // Draw changes
    REQUEST_HANDLER(UpdateDraw);
    REQUEST_HANDLER(SetListDraw);
    REQUEST_HANDLER(DeleteAllDraws);

    // Markers definition changes
    REQUEST_HANDLER(AddMarker);
    REQUEST_HANDLER(DeleteMarker);
    REQUEST_HANDLER(MoveMaker);
    REQUEST_HANDLER(EditMarker);

    // Special markers changes
    REQUEST_HANDLER(SetAbsentMarker);
    REQUEST_HANDLER(SinglesManagementChange);

    // Finishing the registration
    REQUEST_HANDLER(FinishRegistration);

    #undef REQUEST_HANDLER

    // These functions have custom signatures and thus can't be generated via REQUEST_HANDLER

    // Ranking comparison
    void processRankingRequest(Connection *connection, const QString &client, int clientId);

    // Score comparison
    void processScoreRequest(int round, Connection *connection, const QString &client,
                             int clientId);

private: // Variables
    const int m_connectTimeout;
    Database *m_db;
    Settings *m_settings;
    TournamentSettings *m_tournamentSettings;
    RegistrationPage *m_registrationPage;
    MarkersWidget *m_markersWidget;
    const QMap<int, QString> *m_scoreChecksums;
    const QString *m_checksum;

    const QString m_locale;
    const QHash<QString, void(Server::*)(const QJsonObject &, const QString &, int)>
              m_requestHandlers;

    QVector<Connection *> m_pendingConnections;
    QVector<Connection *> m_connections;
    int m_clientIdCounter = 0;

};

#endif // SERVER_H
