/* SPDX-FileCopyrightText: 2010-2020 Tobias Leupold <tobias.leupold@gmx.de>

   SPDX-License-Identifier: GPL-3.0-or-later
*/

// Local includes

#include "TournamentInfoDialog.h"

#include "shared/SharedObjects.h"
#include "shared/Database.h"
#include "shared/TournamentSettings.h"

// Qt includes
#include <QGridLayout>
#include <QLabel>
#include <QDir>
#include <QGroupBox>
#include <QDebug>

TournamentInfoDialog::TournamentInfoDialog(QWidget *parent, SharedObjects *sharedObjects)
    : TitleDialog(parent)
{
    // Cache the pointers to the database and the tournament settings for convenience
    Database *db = sharedObjects->database();
    TournamentSettings *ts = sharedObjects->tournamentSettings();
    const bool isPairMode = ts->isPairMode();

    setTitle(tr("Turnierinformationen"));
    addButtonBox(QDialogButtonBox::Ok);

    addGroup(tr("Datenbank"));

    addLabel(tr("Dateiname:"));
    QString dbFile = QDir::toNativeSeparators(db->dbFile());
    QFont font = this->font();
    QFontMetrics fontMetrics(font);
    addInfo(dbFile, fontMetrics.boundingRect(dbFile).width() > fontMetrics.boundingRect(
                        QStringLiteral("/A/very/long/path/which/still/does/not/mess/up/"
                                       "the/layout")).width());

    addLabel(tr("Erstellt mit:"));
    addInfo(tr("Muckturnier %1").arg(ts->muckturnierVersion()));

    addGroup(tr("Turnierdaten"));

    addLabel(tr("Turniermodus:"));
    addInfo(isPairMode ? tr("Feste Paare") : tr("Einzelne Spieler"));

    addLabel(tr("Punkte pro Bobbl:"));
    addInfo(QString::number(ts->boogerScore()));

    addLabel(tr("Bobbl pro Runde:"));
    addInfo(QString::number(ts->boogersPerRound()));

    addLabel(isPairMode ? tr("Angemeldete Paare:") : tr("Angemeldete Spieler:"));
    addInfo(QString::number(db->pairsOrPlayersCount()));

    addGroup(tr("Ergebnisse"));

    if (db->tournamentStarted()) {
        addLabel(tr("Turnierstatus:"));
        addInfo(ts->tournamentOpen() ? tr("Offen") : tr("Abgeschlossen"));

        addLabel(tr("Gespielte Runden:"));
        addInfo(QString::number(db->allRounds()));

        if (ts->tournamentOpen()) {
            addLabel(tr("Alle Runden abgeschlossen:"));
            addInfo(db->allRoundsFinished() ? tr("Ja") : tr("Nein"));
        }

    } else {
        addLabel(tr("<i>Es liegen noch keine Ergebnisse vor</i>"));
    }
}

void TournamentInfoDialog::addGroup(const QString &text)
{
    QGroupBox *group = new QGroupBox(text);
    layout()->addWidget(group);
    QGridLayout *layout = new QGridLayout(group);
    m_currentGroupLayout = layout;
    m_row = -1;
}

void TournamentInfoDialog::addLabel(const QString &text)
{
    m_row++;
    QLabel *label = new QLabel(text);
    label->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Minimum);
    label->setTextInteractionFlags(Qt::TextBrowserInteraction | Qt::TextSelectableByMouse);
    label->setAlignment(Qt::AlignTop);
    m_currentGroupLayout->addWidget(label, m_row, 0);
}

void TournamentInfoDialog::addInfo(const QString &text, bool wordWrap)
{
    QLabel *label = new QLabel(text);
    label->setTextInteractionFlags(Qt::TextBrowserInteraction | Qt::TextSelectableByMouse);
    if (wordWrap) {
        label->setWordWrap(true);
    }
    m_currentGroupLayout->addWidget(label, m_row, 1);
}
