/* SPDX-FileCopyrightText: 2018-2020 Tobias Leupold <tobias.leupold@gmx.de>

   SPDX-License-Identifier: GPL-3.0-or-later
*/

#ifndef SERVER_PROTOCOL_H
#define SERVER_PROTOCOL_H

#include <QJsonObject>
#include <QDataStream>

namespace ServerProtocol
{

// Basic protocol data
constexpr int protocolVersion = 3;
const QDataStream::Version streamVersion = QDataStream::Qt_5_11;
const QLatin1String serverSignature("muckturnier server");

// Keys used in all requests and answers
const QLatin1String action  ("action");
const QLatin1String clientId("client id");
const QLatin1String checksum("checksum");

// Invalid request handling

const QLatin1String actionRejected("request rejected");
const QLatin1String rejectReason("reject reason");

enum RejectReason {
    Unknown,
    OutdatedRequest,
    UnknownAction
};

// Data types
// ==========

// Markers
// -------

const QLatin1String markerName     ("marker name");
const QLatin1String markerId       ("marker id");
const QLatin1String markerColor    ("marker color");
const QLatin1String markerSorting  ("marker sorting");
const QLatin1String newMarkerName  ("new marker name");
const QLatin1String markerSequence ("marker sequence");
const QLatin1String markerDirection("marker direction");
const QLatin1String absentMarker   ("absent marker");

// Players
// -------

const QLatin1String playersName       ("players name");
const QLatin1String newPlayersName    ("new players name");
const QLatin1String playersTableNumber("players table number");
const QLatin1String playersPairNumber ("players pair number");

// Data lists
// ----------

const QLatin1String playersList       ("players list");
const QLatin1String markersList       ("markers list");
const QLatin1String tablesList        ("tables list");
const QLatin1String pairNumbersList   ("pair numbers list");
const QLatin1String boogersList       ("boogers list");
const QLatin1String goalsList         ("goals list");
const QLatin1String opponentGoalsList ("opponent goals list");
const QLatin1String playedBoogersList ("played boogers list");
const QLatin1String scoresList        ("scores list");
const QLatin1String scoreChecksumsList("score checksums");

// Other (request specific) keys
// -----------------------------

const QLatin1String singlesManagement("singles management");
const QLatin1String rankingAvailable ("ranking available");
const QLatin1String round            ("round");
const QLatin1String scoreChecksum    ("score checksum");

// Handshake
// =========

const QLatin1String handshake("handshake");
const QLatin1String phase    ("phase");
const QLatin1String failed   ("failed");

// Welcome message
const QLatin1String initialization ("initialization");
const QLatin1String serverSignatureKey("server signature");
const QLatin1String protocolVersionKey("protocol version");
const QJsonObject welcomeMessage {
    { action,             handshake },
    { phase,              initialization },
    { serverSignatureKey, serverSignature },
    { protocolVersionKey, protocolVersion }
};

// Synchronization request
const QLatin1String synchronization("synchronization");
const QJsonObject synchronizationRequest {
    { action, handshake },
    { phase,  synchronization }
};

// Synchronization answer

const QLatin1String settings         ("settings");
const QLatin1String locale           ("locale");
const QLatin1String namesSeparator   ("names separator");
const QLatin1String tournamentMode   ("tournament mode");
const QLatin1String boogerScore      ("booger score");
const QLatin1String boogersPerRound  ("boogers per round");
const QLatin1String tournamentStarted("tournament started");

// Synchronization success, request/allocate an id
const QLatin1String idAllocation("id allocation");
const QJsonObject idAllocationRequest {
    { action, handshake },
    { phase,  idAllocation }
};

// Actions
// =======

// Registration changes

const QLatin1String registerPlayers    ("register players");
const QLatin1String renamePlayers      ("rename players");
const QLatin1String assemblePair       ("assemble pair");
const QLatin1String deletePlayers      ("delete players");
const QLatin1String updatePlayersMarker("update players marker");

const QLatin1String deleteMarkedPlayers  ("delete marked players");
const QLatin1String deleteUnmarkedPlayers("delete unmarked players");
const QLatin1String deleteAllPlayers     ("delete all players");

const QLatin1String updatePlayersListMarker("update players list marker");
const QLatin1String removePlayersListMarker("remove players list marker");

// Draw changes
const QLatin1String updateDraw    ("update draw");
const QLatin1String setListDraw   ("set list draw");
const QLatin1String deleteAllDraws("delete all draws");

// Markers changes
const QLatin1String addMarker   ("add marker");
const QLatin1String editMarker  ("edit marker");
const QLatin1String deleteMarker("delete marker");
const QLatin1String moveMarker  ("move marker");

// Special markers definitions
const QLatin1String setAbsentMarker        ("set absent marker");
const QLatin1String singlesManagementChange("singles management change");

// Finish registration
const QLatin1String finishRegistration("finish registration");

// Score change
const QLatin1String scoreChange("score change");

// Score request
const QLatin1String scoreRequest("score request");

// Ranking request
const QLatin1String rankingRequest("ranking request");

}

#endif // SERVER_PROTOCOL_H
