/* SPDX-FileCopyrightText: 2019-2020 Tobias Leupold <tobias.leupold@gmx.de>

   SPDX-License-Identifier: GPL-3.0-or-later
*/

// Local includes
#include "JsonHelper.h"

// Qt includes
#include <QString>
#include <QJsonObject>
#include <QJsonArray>
#include <QJsonDocument>
#include <QDebug>

namespace JsonHelper
{

// ===========================================================================================
// If anything is changed in the values of the QStringLiterals here,
// don't forget to bump the server protocol version!
static const QString s_singlesManagement_enabled         = QStringLiteral("enabled");
static const QString s_singlesManagement_singleMarker    = QStringLiteral("single marker");
static const QString s_singlesManagement_markSingles     = QStringLiteral("mark singles");
static const QString s_singlesManagement_conditionString = QStringLiteral("condition string");
static const QString s_singlesManagement_condition       = QStringLiteral("condition");
static const QString s_singlesManagement_assignedMarker  = QStringLiteral("assigned marker");
// ===========================================================================================

static const QString s_draw_consecutive     = QStringLiteral("consecutive");
static const QString s_draw_tablesLimit     = QStringLiteral("tables limit");
static const QString s_draw_randomDraw      = QStringLiteral("random draw");
static const QString s_draw_windowDraw      = QStringLiteral("window draw");
static const QString s_draw_consecutiveDraw = QStringLiteral("consecutive draw");

QString serializeJson(const QJsonObject &json)
{
    return QString::fromUtf8(QJsonDocument(json).toJson(QJsonDocument::Compact));
}

QString serializeJson(const QJsonArray &json)
{
    return QString::fromUtf8(QJsonDocument(json).toJson(QJsonDocument::Compact));
}

QJsonObject jsonObjectFromString(const QString &data)
{
    return QJsonDocument::fromJson(data.toUtf8()).object();
}

QJsonArray jsonArrayFromString(const QString &data)
{
    return QJsonDocument::fromJson(data.toUtf8()).array();
}

QJsonObject singlesManagementToJson(const Markers::SinglesManagement &data)
{
    return QJsonObject {
        { s_singlesManagement_enabled,         data.enabled },
        { s_singlesManagement_singleMarker,    data.singleMarker },
        { s_singlesManagement_markSingles,     data.markSingles },
        { s_singlesManagement_conditionString, data.conditionString },
        { s_singlesManagement_condition,       data.condition },
        { s_singlesManagement_assignedMarker,  data.assignedMarker }
    };
}

QString singlesManagementToString(const Markers::SinglesManagement &data)
{
    return QString::fromUtf8(
        QJsonDocument(singlesManagementToJson(data)).toJson(QJsonDocument::Compact));
}

Markers::SinglesManagement parseSinglesManagement(const QString &data)
{
    return parseSinglesManagement(jsonObjectFromString(data));
}

Markers::SinglesManagement parseSinglesManagement(const QJsonObject &jsonObject)
{
    Markers::SinglesManagement data;
    QJsonValue value;

    value = jsonObject.value(s_singlesManagement_enabled);
    data.enabled = value.type() == QJsonValue::Bool
        ? value.toBool() : Markers::defaultSinglesManagement.enabled;

    value = jsonObject.value(s_singlesManagement_singleMarker);
    data.singleMarker = value.type() == QJsonValue::Double
        ? value.toInt() : Markers::defaultSinglesManagement.singleMarker;

    value = jsonObject.value(s_singlesManagement_markSingles);
    data.markSingles = value.type() == QJsonValue::Bool
        ? value.toBool() : Markers::defaultSinglesManagement.markSingles;

    value = jsonObject.value(s_singlesManagement_conditionString);
    data.conditionString = value.type() == QJsonValue::String
        ? value.toString() : Markers::defaultSinglesManagement.conditionString;

    value = jsonObject.value(s_singlesManagement_condition);
    data.condition = value.type() == QJsonValue::Double
        ? value.toInt() : Markers::defaultSinglesManagement.condition;

    value = jsonObject.value(s_singlesManagement_assignedMarker);
    data.assignedMarker = value.type() == QJsonValue::Double
        ? value.toInt() : Markers::defaultSinglesManagement.assignedMarker;

    return data;
}

QString registrationColumnsToString(const QVector<RegistrationColumns::Column> &columns)
{
    return serializeJson(registrationColumnsToJson(columns));
}

QJsonArray registrationColumnsToJson(const QVector<RegistrationColumns::Column> &columns)
{
    QJsonArray data;
    for (const RegistrationColumns::Column column : RegistrationColumns::allColumns) {
        if (columns.contains(column)) {
            data.append(RegistrationColumns::columnId.value(column));
        }
    }
    return data;
}

QVector<RegistrationColumns::Column> parseRegistrationColumns(const QString &data)
{
    if (! data.isEmpty()) {
        return parseRegistrationColumns(jsonArrayFromString(data));
    } else {
        return RegistrationColumns::defaultColumns;
    }
}

QVector<RegistrationColumns::Column> parseRegistrationColumns(const QJsonArray &jsonArray)
{
    QVector<RegistrationColumns::Column> columns;
    for (const RegistrationColumns::Column column : RegistrationColumns::allColumns) {
        if (jsonArray.contains(RegistrationColumns::columnId.value(column))) {
            columns.append(column);
        }
    }
    return columns;
}

QString drawSettingsToString(const Draw::Settings &settings)
{
    return serializeJson(drawSettingsToJson(settings));
}

QJsonObject drawSettingsToJson(const Draw::Settings &settings)
{
    return QJsonObject {
        { s_draw_consecutive,     settings.consecutive },
        { s_draw_tablesLimit,     settings.tablesLimit },
        { s_draw_randomDraw,      settings.randomDraw },
        { s_draw_windowDraw,      settings.windowDraw },
        { s_draw_consecutiveDraw, settings.consecutiveDraw }
    };
}

Draw::Settings parseDrawSettings(const QString &data)
{
    return parseDrawSettings(jsonObjectFromString(data));
}

Draw::Settings parseDrawSettings(const QJsonObject &jsonObject)
{
    const bool consecutive = jsonObject.value(s_draw_consecutive).toBool();
    const int tablesLimit = jsonObject.value(s_draw_tablesLimit).toInt();
    const int randomDraw = jsonObject.value(s_draw_randomDraw).toInt();
    const int windowDraw = jsonObject.value(s_draw_windowDraw).toInt();
    const int consecutiveDraw = jsonObject.value(s_draw_consecutiveDraw).toInt();

    if (! consecutive
        && ((tablesLimit != -1 && (randomDraw > tablesLimit || consecutiveDraw > tablesLimit))
            || randomDraw >= consecutiveDraw
            || (randomDraw != -1 && consecutiveDraw != -1
                && consecutiveDraw == randomDraw + 1
                && windowDraw != -1))) {

        return Draw::defaultSettings;
    }

    return Draw::Settings {
        consecutive,
        tablesLimit,
        randomDraw,
        windowDraw,
        consecutiveDraw
    };
}

const QString OPTIONAL_PAGES_DRAW_OVERVIEW = QStringLiteral("draw_overview");

const QHash<OptionalPages::Page, QString>  OPTIONAL_PAGES_ID {
    { OptionalPages::Page::DrawOverview, OPTIONAL_PAGES_DRAW_OVERVIEW }
};

const QHash<QString, OptionalPages::Page>  OPTIONAL_PAGES_ID_MAP {
    { OPTIONAL_PAGES_DRAW_OVERVIEW, OptionalPages::Page::DrawOverview }
};

QString optionalPagesToString(const QVector<OptionalPages::Page> &pages)
{
    return serializeJson(optionalPagesToJson(pages));
}

QJsonArray optionalPagesToJson(const QVector<OptionalPages::Page> &pages)
{
    QJsonArray ids;
    for (const OptionalPages::Page page : pages) {
        ids.append(OPTIONAL_PAGES_ID.value(page));
    }
    return ids;
}

QVector<OptionalPages::Page> parseOptionalPages(const QString &data)
{
    return parseOptionalPages(jsonArrayFromString(data));
}

QVector<OptionalPages::Page> parseOptionalPages(const QJsonArray &json)
{
    QVector<OptionalPages::Page> pages;
    for (const QJsonValue &value : json) {
        const QString id = value.toString();
        if (OPTIONAL_PAGES_ID_MAP.contains(id)) {
            pages.append(OPTIONAL_PAGES_ID_MAP.value(id));
        }
    }
    return pages;
}

// Legacy functions needed for updates
// ===================================

QVector<RegistrationColumns::Column> parsePlayersPageColumns(const QString &data)
{
    if (data.isEmpty()) {
        return RegistrationColumns::defaultColumns;
    } else {
        return parsePlayersPageColumns(jsonObjectFromString(data));
    }
}

QVector<RegistrationColumns::Column> parsePlayersPageColumns(const QJsonObject &jsonObject)
{
    QVector<RegistrationColumns::Column> columns;
    for (const RegistrationColumns::Column column : RegistrationColumns::allColumns) {
        QJsonValue jsonValue = jsonObject.value(RegistrationColumns::columnId.value(column));
        const bool visible = jsonValue.type() == QJsonValue::Bool
                                 ? jsonValue.toBool()
                                 : RegistrationColumns::defaultColumns.contains(column);
        if (visible) {
            columns.append(column);
        }
    }
    return columns;
}

}
