/* SPDX-FileCopyrightText: 2019-2020 Tobias Leupold <tobias.leupold@gmx.de>

   SPDX-License-Identifier: GPL-3.0-or-later
*/

// Local includes
#include "SelectionPopup.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QCursor>
#include <QScrollArea>
#include <QScrollBar>
#include <QApplication>
#include <QTimer>
#include <QScreen>

SelectionPopup::SelectionPopup(QWidget *parent) : QDialog(parent)
{
    setWindowFlags(Qt::FramelessWindowHint | Qt::Popup);
    setStyleSheet(QStringLiteral(
        "QDialog { border: 1px solid %1; } "
        "QScrollArea { border: none; } "
        "QPushButton { border: none; border-radius: 4px; padding: 0.5em; } "
        "QPushButton:hover { border: 1px solid palette(highlight); } "
    ).arg(qApp->property("DARK_MODE").toBool() ? QStringLiteral("palette(light)")
                                               : QStringLiteral("palette(mid)")));

    auto *layout = new QVBoxLayout(this);
    layout->setSpacing(0);
    layout->setMargin(1); // We need this for the QDialog's gray border

    m_scrollArea = new QScrollArea(this);
    m_scrollArea->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    layout->addWidget(m_scrollArea);
}

void SelectionPopup::setScrollWidget(QWidget *widget)
{
    widget->setStyleSheet(QStringLiteral("QWidget { background-color: palette(base); }"));
    m_scrollArea->setWidget(widget);
}

void SelectionPopup::show(QPoint position)
{
    if (position == QPoint(-1, -1)) {
        position = QCursor::pos();
    }

    const QRect availableGeometry = QGuiApplication::primaryScreen()->availableGeometry();
    const int availableHeight = availableGeometry.height() - position.y();
    const int availableWidth = availableGeometry.width() - position.x();

    const QSize widgetSize = m_scrollArea->widget()->size();
    const int widgetHeight = widgetSize.height() + 2; // Two pixels for the border
    const int widgetWidth = widgetSize.width(); // Who appararenly aren't necessary here

    if (availableWidth < widgetWidth) {
        if (availableGeometry.width() >= widgetWidth) {
            position.setX(position.x() - (widgetWidth - availableWidth));
        } else {
            position.setX(0);
        }
    }

    int targetHeight = -1;

    if (widgetHeight <= availableHeight) {
        targetHeight = widgetHeight;
    } else {
        if (availableGeometry.height() >= widgetHeight) {
            position.setY(position.y() - (widgetHeight - availableHeight));
            targetHeight = widgetHeight;
        } else {
            position.setY(0);
            targetHeight = availableGeometry.height();
        }
    }

    resize(m_scrollArea->widget()->width(), targetHeight);
    move(position);

    QDialog::show();
    QTimer::singleShot(0, this, &SelectionPopup::checkForScrollBar);
}

void SelectionPopup::checkForScrollBar()
{
    if (m_scrollArea->verticalScrollBar()->isVisible()) {
        resize(m_scrollArea->widget()->width() + m_scrollArea->verticalScrollBar()->width(),
               height());
    } else {
        resize(m_scrollArea->widget()->width(), height());
    }
}
