/* SPDX-FileCopyrightText: 2019-2020 Tobias Leupold <tobias.leupold@gmx.de>

   SPDX-License-Identifier: GPL-3.0-or-later
*/

// Local includes
#include "Settings.h"
#include "DefaultValues.h"
#include "JsonHelper.h"
#include "OptionalPages.h"

// Qt includes
#include <QDebug>
#include <QAction>
#include <QJsonObject>
#include <QJsonArray>
#include <QStandardPaths>
#include <QDir>

// C++ includes
#include <utility>

// Config file string literals
// ===========================

// Options check state group
static const QLatin1String s_saveOnClose("save_on_close/");
static const QString s_saveOnClose_mainWindowGeometry
    = s_saveOnClose + QLatin1String("mainwindow_geometry");
static const QString s_saveOnClose_mainWindowState
    = s_saveOnClose + QLatin1String("mainwindow_state");
static const QString s_saveOnClose_lastDbPath
    = s_saveOnClose + QLatin1String("last_db_path");
static const QString s_saveOnClose_stopWatchState
    = s_saveOnClose + QLatin1String("stopwatch_state");
static const QString s_saveOnClose_phoneticSearch
    = s_saveOnClose + QLatin1String("phonetic_search");

// Session data
static const QString s_sessionData = QStringLiteral("session_data/");
static const QString s_sessionData_tabPosition
    = s_sessionData + QLatin1String("tab_position");
static const QString s_sessionData_mainWindowGeometry
    = s_sessionData + QLatin1String("mainwindow_geometry");
static const QString s_sessionData_mainWindowState
    = s_sessionData + QLatin1String("mainwindow_state");
static const QString s_sessionData_lastDbPath
    = s_sessionData + QLatin1String("last_db_path");
static const QString s_sessionData_newTournamentShowExtendedOptions
    = s_sessionData + QLatin1String("new_tournament_show_extended_options");

// Stop watch state
static const QLatin1String s_stopWatchState("stopwatch_state/");
static const QString s_stopWatchState_version
    = s_stopWatchState + QLatin1String("version");
static const QString s_stopWatchState_hours
    = s_stopWatchState + QLatin1String("hours");
static const QString s_stopWatchState_minutes
    = s_stopWatchState + QLatin1String("minutes");
static const QString s_stopWatchState_seconds
    = s_stopWatchState + QLatin1String("seconds");
static const QString s_stopWatchState_title
    = s_stopWatchState + QLatin1String("title");
static const QString s_stopWatchState_geometry
    = s_stopWatchState + QLatin1String("geometry");

// Phonetic search state
static const QString s_phoneticSearch = QStringLiteral("phonetic_search");

// Strings
static const QLatin1String s_strings("strings/");
static const QString s_strings_boogerSymbol = s_strings + QLatin1String("booger_symbol");
static const QString s_strings_namesSeparator = s_strings + QLatin1String("names_separator");

// Tournament template

static const QString s_tournamentTemplate = QStringLiteral("tournament_template");
static const QString s_tournamentTemplate_version
    = s_tournamentTemplate + QLatin1String("/version");
static const QString s_tournamentTemplate_dbFileNameTemplate
    = s_tournamentTemplate + QLatin1String("/db_file_name_template");
static const QString s_tournamentTemplate_tournamentMode
    = s_tournamentTemplate + QLatin1String("/tournament_mode");
static const QString s_tournamentTemplate_scoreType
    = s_tournamentTemplate + QLatin1String("/score_type");
static const QString s_tournamentTemplate_boogerScore
    = s_tournamentTemplate + QLatin1String("/booger_score");
static const QString s_tournamentTemplate_boogersPerRound
    = s_tournamentTemplate + QLatin1String("/boogers_per_round");
static const QString s_tournamentTemplate_registrationColumns
    = s_tournamentTemplate + QLatin1String("/registration_columns");
static const QString s_tournamentTemplate_registrationListHeaderHidden
    = s_tournamentTemplate + QLatin1String("/registration_list_header_hidden");
static const QString s_tournamentTemplate_autoselectPairs
    = s_tournamentTemplate + QLatin1String("/auto_select_pairs");
static const QString s_tournamentTemplate_selectByLastRound
    = s_tournamentTemplate + QLatin1String("/select_by_last_round");
static const QString s_tournamentTemplate_noTableNumbers
    = s_tournamentTemplate + QLatin1String("/no_table_numbers");
static const QString s_tournamentTemplate_includeOpponentGoals
    = s_tournamentTemplate + QLatin1String("/include_opponent_goals");
static const QString s_tournamentTemplate_grayOutUnimportantGoals
    = s_tournamentTemplate + QLatin1String("/gray_out_unimportant_goals");
static const QString s_tournamentTemplate_optionalPages
    = s_tournamentTemplate + QLatin1String("/optional_pages");

// Markers template

static constexpr int s_markersTemplate_version = 2;

static const QString s_markersTemplate = QStringLiteral("markers_template");

static const QString s_markersTemplate_versionKey = s_markersTemplate + QLatin1String("/version");
static const QString s_markersTemplate_markers = s_markersTemplate + QLatin1String("/markers");
static const QLatin1String s_markersTemplate_nameKey   ("n");
static const QLatin1String s_markersTemplate_colorKey  ("c");
static const QLatin1String s_markersTemplate_sortingKey("s");

static const QHash<Markers::Type, QString> s_markersTemplate_markerKey {
    { Markers::Default, s_markersTemplate + QLatin1String("/default_marker") },
    { Markers::Drawn,   s_markersTemplate + QLatin1String("/drawn_marker") },
    { Markers::Absent,  s_markersTemplate + QLatin1String("/absent_marker") }
};

static const QHash<Markers::Type, int> s_markersTemplate_markerDefaultValue {
    { Markers::Default,  0 },
    { Markers::Drawn,   -1 },
    { Markers::Absent,  -1 }
};

// Singles Management

static const QString s_singlesManagementTemplate = QStringLiteral("singles_management_template");

static const QString s_singlesManagementTemplate_version
    = s_singlesManagementTemplate + QLatin1String("/version");
static const QString s_singlesManagementTemplate_data
    = s_singlesManagementTemplate + QLatin1String("/data");

// Network settings

static const QString s_network = QStringLiteral("network");

static const QString s_network_preferIpv6       = s_network + QLatin1String("/prefer_ipv6");

static const QString s_network_serverPort       = s_network + QLatin1String("/server_port");
static const QString s_network_connectTimeout   = s_network + QLatin1String("/connect_timeout");
static const QString s_network_requestTimeout   = s_network + QLatin1String("/request_timeout");

static const QString s_network_discoverPort     = s_network + QLatin1String("/discover_port");
static const QString s_network_discoverTimeout  = s_network + QLatin1String("/discover_timeout");
static const QString s_network_discoverAttempts = s_network + QLatin1String("/discover_attempts");

// QTabWidget::TabPosition map
static const QVector<QString> s_possibleTabPositions {
    QStringLiteral("north"),
    QStringLiteral("south"),
    QStringLiteral("west"),
    QStringLiteral("east")
};

// Draw settings
static const QString s_drawSettingsTemplate = QStringLiteral("draw_settings_template");
static const QString s_drawSettingsTemplate_version
    = s_drawSettingsTemplate + QLatin1String("/version");
static const QString s_drawSettingsTemplate_data
    = s_drawSettingsTemplate + QLatin1String("/data");

// Shared strings
// ==============

const QString Settings::registrationPhoneticId = QStringLiteral("registration_page_register");

// Settings functions
// ==================

Settings::Settings(QObject *parent)
    : QSettings(QStringLiteral("muckturnier.org"), QStringLiteral("muckturnier"), parent)
{
    m_boogerSymbol = value(s_strings_boogerSymbol, DefaultValues::boogerSymbol).toString();
    if (m_boogerSymbol.left(1).simplified().isEmpty()) {
        m_boogerSymbol = DefaultValues::boogerSymbol;
    }

    m_namesSeparator = value(s_strings_namesSeparator, DefaultValues::namesSeparator).toString();
    m_namesSeparator.remove(QLatin1Char('\t'));
}

void Settings::setSaveOnCloseValue(const QString &key, const QString &dataSet, bool state)
{
    setValue(key, state);
    if (! state) {
        remove(dataSet);
    }
}

bool Settings::saveOnCloseValue(const QString &key) const
{
    const QVariant data = value(key);
    return data.isValid() ? data.toBool() : true;
}

// Tab position
// ------------

void Settings::setTabPosition(const QString &position)
{
    setValue(s_sessionData_tabPosition, position);
}

QString Settings::tabPosition() const
{
    const QString position = value(s_sessionData_tabPosition, QStringLiteral("south")).toString();
    if (s_possibleTabPositions.contains(position)) {
        return position;
    } else {
        return QStringLiteral("south");
    }
}

// Main window geometry
// --------------------

bool Settings::saveMainWindowGeometryEnabled() const
{
    return saveOnCloseValue(s_saveOnClose_mainWindowGeometry);
}

void Settings::enableSaveMainWindowGeometry(bool state)
{
    setSaveOnCloseValue(s_saveOnClose_mainWindowGeometry, s_sessionData_mainWindowGeometry, state);
}

void Settings::saveMainWindowGeometry(const QByteArray &data)
{
    setValue(s_sessionData_mainWindowGeometry, data);
}

QByteArray Settings::mainWindowGeometry() const
{
    return value(s_sessionData_mainWindowGeometry, QByteArray()).toByteArray();
}

// Main window state
// -----------------

bool Settings::saveMainWindowStateEnabled() const
{
    return saveOnCloseValue(s_saveOnClose_mainWindowState);
}

void Settings::enableSaveMainWindowState(bool state)
{
    setSaveOnCloseValue(s_saveOnClose_mainWindowState, s_sessionData_mainWindowState, state);
}

void Settings::saveMainWindowState(const QByteArray &data)
{
    if (! data.isEmpty()) {
        setValue(s_sessionData_mainWindowState, data);
    } else {
        remove(s_sessionData_mainWindowState);
    }
}

QByteArray Settings::mainWindowState() const
{
    return value(s_sessionData_mainWindowState, QByteArray()).toByteArray();
}

// Last database path
// ------------------

bool Settings::saveLastDbPathEnabled() const
{
    return saveOnCloseValue(s_saveOnClose_lastDbPath);
}

void Settings::enableSaveLastDbPath(bool state)
{
    setSaveOnCloseValue(s_saveOnClose_lastDbPath, s_sessionData_lastDbPath, state);
}

void Settings::saveLastDbPath(const QString &dbPath)
{
    setValue(s_sessionData_lastDbPath, dbPath);
}

QString Settings::lastDbPath() const
{
    const QString documentsLocation
        = QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation);
    const QString path = value(s_sessionData_lastDbPath, documentsLocation).toString();
    return QDir(path).exists() ? path : documentsLocation;
}

// Stop watch state
// ----------------

void Settings::enableSaveStopWatchState(bool state)
{
    setSaveOnCloseValue(s_saveOnClose_stopWatchState, s_stopWatchState, state);
}

bool Settings::saveStopWatchStateEnabled() const
{
    return saveOnCloseValue(s_saveOnClose_stopWatchState);
}

void Settings::saveStopWatchState(StopWatchState::State state)
{
    setValue(s_stopWatchState_version,  StopWatchState::version);
    setValue(s_stopWatchState_hours,    state.hours);
    setValue(s_stopWatchState_minutes,  state.minutes);
    setValue(s_stopWatchState_seconds,  state.seconds);
    setValue(s_stopWatchState_title,    state.title);
    setValue(s_stopWatchState_geometry, state.geometry);
}

StopWatchState::State Settings::stopWatchState() const
{
    if (! saveStopWatchStateEnabled()
        || value(s_stopWatchState_version, 0).toInt() != StopWatchState::version) {

        return StopWatchState::defaultState;
    }

    int hours = value(s_stopWatchState_hours, 0).toInt();
    if (hours < 0 || hours > 99) {
        hours = 0;
    }

    int minutes = value(s_stopWatchState_minutes, 0).toInt();
    if (minutes < 0 || minutes > 59) {
        minutes = 0;
    }

    int seconds = value(s_stopWatchState_seconds, 0).toInt();
    if (seconds < 0 || seconds > 59) {
        seconds = 0;
    }

    return StopWatchState::State {
        hours,
        minutes,
        seconds,
        value(s_stopWatchState_title, 0).toString(),
        value(s_stopWatchState_geometry, QByteArray()).toByteArray()
    };
}

// "Extended options" of the "New tournament" dialog

bool Settings::newTournamentShowExtendedOptions() const
{
    return value(s_sessionData_newTournamentShowExtendedOptions, false).toBool();
}

void Settings::setNewTournamentShowExtendedOptions(bool state)
{
    setValue(s_sessionData_newTournamentShowExtendedOptions, state);
}

// Phonetic search state

void Settings::enableSavePhoneticSearchState(bool state)
{
    setSaveOnCloseValue(s_saveOnClose_phoneticSearch, s_phoneticSearch, state);
}

bool Settings::savePhoneticSearchStateEnabled() const
{
    return saveOnCloseValue(s_saveOnClose_phoneticSearch);
}

bool Settings::phoneticSearchState(const QString &searchId) const
{
    const bool defaultValue = searchId == Settings::registrationPhoneticId
                                  ? true : false;
    return value(s_phoneticSearch + QStringLiteral("/") + searchId, defaultValue).toBool();
}

void Settings::savePhoneticSearchState(const QString &searchId, bool state)
{
    if (savePhoneticSearchStateEnabled()) {
        setValue(s_phoneticSearch + QStringLiteral("/") + searchId, state);
    }
}

// Tournament template
// -------------------

void Settings::saveTournamentTemplate(const TournamentTemplate::Settings &tournamentTemplate)
{
    const int version = value(s_tournamentTemplate_version, -1).toInt();
    if (version < 3) {
        // Clean up the old "players page" category
        const QString playersPage = QStringLiteral("players_page");

        // The "columns" key has been moved to "tournament_template/players_page_columns"
        remove(playersPage + QLatin1String("/columns"));

        // Check if there are other keys in "players_page" (maybe added by later versions)
        beginGroup(playersPage);
        const bool otherKeysExist = allKeys().count() > 0;
        endGroup();

        // If not, remove the "players_page" category
        if (! otherKeysExist) {
            remove(playersPage);
        }
    }

    setValue(s_tournamentTemplate_version, TournamentTemplate::version);
    setValue(s_tournamentTemplate_tournamentMode,
             Tournament::modeMap.value(tournamentTemplate.tournamentMode));
    setValue(s_tournamentTemplate_scoreType,
             Tournament::scoreTypeMap.value(tournamentTemplate.scoreType));
    setValue(s_tournamentTemplate_boogerScore, tournamentTemplate.boogerScore);
    setValue(s_tournamentTemplate_boogersPerRound, tournamentTemplate.boogersPerRound);
    setValue(s_tournamentTemplate_registrationColumns,
             JsonHelper::registrationColumnsToString(tournamentTemplate.registrationColumns));
    setValue(s_tournamentTemplate_registrationListHeaderHidden,
             tournamentTemplate.registrationListHeaderHidden);
    setValue(s_tournamentTemplate_autoselectPairs, tournamentTemplate.autoSelectPairs);
    setValue(s_tournamentTemplate_selectByLastRound, tournamentTemplate.selectByLastRound);
    setValue(s_tournamentTemplate_noTableNumbers, tournamentTemplate.noTableNumbers);
    setValue(s_tournamentTemplate_includeOpponentGoals, tournamentTemplate.includeOpponentGoals);
    setValue(s_tournamentTemplate_grayOutUnimportantGoals,
             tournamentTemplate.grayOutUnimportantGoals);
    setValue(s_tournamentTemplate_optionalPages,
             JsonHelper::optionalPagesToString(tournamentTemplate.optionalPages));

    saveDrawSettingsTemplate(tournamentTemplate.drawSettings);
}

TournamentTemplate::Settings Settings::tournamentTemplate() const
{
    const int version = value(s_tournamentTemplate_version, -1).toInt();
    if (version <= 0 || version > TournamentTemplate::version) {
        return TournamentTemplate::defaultSettings;
    }

    // Get the tournament mode
    Tournament::Mode tournamentMode;
    if (version >= 2) {
        // Check for a correct value of s_tournamentTemplate_tournamentMode
        const QString tournamentModeString
            = value(s_tournamentTemplate_tournamentMode,
                    Tournament::fixedPairsMode).toString();
        if (! Tournament::modeMap.values().contains(tournamentModeString)) {
            return TournamentTemplate::defaultSettings;
        } else {
            tournamentMode = Tournament::modeMap.key(tournamentModeString);
        }
    } else if (version < 2) {
        // Read the old bool "pair_mode" value
        tournamentMode = value(s_tournamentTemplate + QLatin1String("/pair_mode"), true).toBool()
                             ? Tournament::Mode::FixedPairs : Tournament::Mode::SinglePlayers;
    }

    // Get the score type
    Tournament::ScoreType scoreType;
    const QString scoreTypeString = value(s_tournamentTemplate_scoreType,
                                          Tournament::horizontalScoreType).toString();
    if (! Tournament::scoreTypeMap.values().contains(scoreTypeString)) {
        return TournamentTemplate::defaultSettings;
    } else {
        scoreType = Tournament::scoreTypeMap.key(scoreTypeString);
    }

    // Check for correct booger score and boogers per round
    const int boogerScore = value(s_tournamentTemplate_boogerScore,
                                  DefaultValues::boogerScore).toInt();
    const int boogersPerRound = value(s_tournamentTemplate_boogersPerRound,
                                      DefaultValues::boogersPerRound).toInt();
    if (boogerScore < 1 || boogersPerRound < 1) {
        return TournamentTemplate::defaultSettings;
    }

    // Get the players page columns
    QVector<RegistrationColumns::Column> registrationColumns;
    if (version >= 3) {
        registrationColumns = JsonHelper::parseRegistrationColumns(
                                  value(s_tournamentTemplate_registrationColumns).toString());
    } else if (version < 3) {
        registrationColumns = JsonHelper::parsePlayersPageColumns(
                                  value(QStringLiteral("players_page/columns")).toString());
    }

    // Assemble the template and return it
    return TournamentTemplate::Settings {
        tournamentMode,
        scoreType,
        boogerScore,
        boogersPerRound,
        drawSettingsTemplate(),
        registrationColumns,
        value(s_tournamentTemplate_registrationListHeaderHidden, false).toBool(),
        value(s_tournamentTemplate_autoselectPairs, true).toBool(),
        value(s_tournamentTemplate_selectByLastRound, true).toBool(),
        value(s_tournamentTemplate_noTableNumbers, false).toBool(),
        value(s_tournamentTemplate_includeOpponentGoals, true).toBool(),
        value(s_tournamentTemplate_grayOutUnimportantGoals, true).toBool(),
        JsonHelper::parseOptionalPages(value(s_tournamentTemplate_optionalPages).toString())
    };
}

// Database file name template
// ---------------------------

void Settings::saveDbFileNameTemplate(const QString &text)
{
    setValue(s_tournamentTemplate_dbFileNameTemplate, text);
}

QString Settings::dbFileNameTemplate() const
{
    return value(s_tournamentTemplate_dbFileNameTemplate,
                 DefaultValues::dbFileNameTemplate).toString();
}

// Markers template
// ----------------

void Settings::saveMarkersTemplate(const QVector<Markers::Marker> &markers)
{
    QJsonArray json;
    for (const Markers::Marker &marker : markers) {
        json.append(QJsonObject {
            { s_markersTemplate_nameKey,    marker.name },
            { s_markersTemplate_colorKey,   QString::number(marker.color.rgb()) },
            { s_markersTemplate_sortingKey, static_cast<int>(marker.sorting) } });
    }

    setValue(s_markersTemplate_versionKey, s_markersTemplate_version);
    setValue(s_markersTemplate_markers, JsonHelper::serializeJson(json));
}

bool Settings::parseMarkerData(int &id, const QJsonValue &data,
                               QVector<Markers::Marker> &markers) const
{
    Markers::Marker marker;
    const QJsonObject markerData = data.toObject();
    QJsonValue value;

    value = markerData.value(s_markersTemplate_nameKey);
    if (value.type() != QJsonValue::String) {
        return false;
    } else {
        marker.name = value.toString();
    }

    value = markerData.value(s_markersTemplate_colorKey);
    if (value.type() != QJsonValue::String) {
        return false;
    } else {
        marker.color = QColor::fromRgb(value.toString().toUInt());
    }

    value = markerData.value(s_markersTemplate_sortingKey);
    if (value.type() != QJsonValue::Double) {
        return false;
    } else {
        marker.sorting = static_cast<Markers::MarkerSorting>(value.toInt());
    }

    if (marker.name.isEmpty()) {
        marker.id = 0;
    } else {
        marker.id = ++id;
    }
    markers.append(marker);

    return true;
}

void Settings::appendMarker(const QString &name, const QColor &color,
                            Markers::MarkerSorting sorting, int id,
                            QVector<Markers::Marker> &markers) const
{
    Markers::Marker marker;
    marker.name = name;
    marker.color = color;
    marker.sorting = sorting;
    marker.id = id;
    markers.append(marker);
}

QVector<Markers::Marker> Settings::defaultMarkersTemplate() const
{
    QVector<Markers::Marker> markers;
    markers.reserve(4);
    appendMarker(tr("Rot"),  QColor(255, 0,   0,   255),
                 Markers::MarkerSorting::NewBlock, 1, markers);
    appendMarker(tr("Blau"), QColor(0,   0,   255, 255),
                 Markers::MarkerSorting::NewBlock, 2, markers);
    appendMarker(QString(),  QColor(0,   0,   0,   255),
                 Markers::MarkerSorting::NewBlock, 0, markers);
    appendMarker(tr("Grün"), QColor(0,   128, 0,   255),
                 Markers::MarkerSorting::Merged,   3, markers);
    return markers;
}

QVector<Markers::Marker> Settings::markersTemplate() const
{
    const int version = value(s_markersTemplate_versionKey, 0).toInt();

    QJsonArray data;
    if (version == 2) {
        data = JsonHelper::jsonArrayFromString(value(s_markersTemplate_markers,
                                                     QString()).toString());
    } else if (version == 1) {
        data = JsonHelper::jsonObjectFromString(QString::fromUtf8(
            value(QStringLiteral("markers_template/data"), QByteArray()).toByteArray())).value(
                QStringLiteral("d")).toArray();
    } else {
        return defaultMarkersTemplate();
    }

    if (data.isEmpty()) {
        return defaultMarkersTemplate();
    }

    QVector<Markers::Marker> markers;

    bool templateParsed = true;
    int id = 0;

    for (const QJsonValue &value : std::as_const(data)) {
        if (! parseMarkerData(id, value, markers)) {
            templateParsed = false;
            break;
        }
    }

    // Be sure to have an "(unmarked)" marker with ID 0 and no name
    if (templateParsed) {
        templateParsed = false;
        for (const Markers::Marker &marker : std::as_const(markers)) {
            if (marker.id == 0 && marker.name.isEmpty()) {
                templateParsed = true;
                break;
            }
        }
    }

    if (! templateParsed) {
        return defaultMarkersTemplate();
    }

    return markers;
}

void Settings::saveMarker(Markers::Type type, int id)
{
    setValue(s_markersTemplate_markerKey.value(type), id);
}

int Settings::marker(Markers::Type type) const
{
    return value(s_markersTemplate_markerKey.value(type),
                 s_markersTemplate_markerDefaultValue.value(type)).toInt();
}

// Singles Management
// ------------------

void Settings::saveSinglesManagementTemplate(const Markers::SinglesManagement &data)
{
    setValue(s_singlesManagementTemplate_version, Markers::singlesManagementVersion);
    setValue(s_singlesManagementTemplate_data,    JsonHelper::singlesManagementToString(data));
}

Markers::SinglesManagement Settings::defaultSinglesManagementTemplate() const
{
    Markers::SinglesManagement data = Markers::defaultSinglesManagement;
    data.conditionString = namesSeparator().simplified();
    return data;
}

Markers::SinglesManagement Settings::singlesManagementTemplate() const
{
    if (value(s_singlesManagementTemplate_version, 0).toInt()
        != Markers::singlesManagementVersion) {

        return defaultSinglesManagementTemplate();
    }

    return JsonHelper::parseSinglesManagement(value(s_singlesManagementTemplate_data).toString());
}

// Strings
// -------

void Settings::saveBoogerSymbol(const QString &text)
{
    setValue(s_strings_boogerSymbol, text);
    m_boogerSymbol = text;
}

const QString &Settings::boogerSymbol() const
{
    return m_boogerSymbol;
}

void Settings::saveNamesSeparator(const QString &text)
{
    setValue(s_strings_namesSeparator, text);
    m_namesSeparator = text;
}

void Settings::setTransientNamesSeparator(const QString &text)
{
    if (m_namesSeparator != text) {
        m_originalNamesSeparator = m_namesSeparator;
        m_namesSeparator = text;
    } else {
        m_originalNamesSeparator.clear();
    }
}

bool Settings::resetNamesSeparator()
{
    if (! m_originalNamesSeparator.isEmpty()) {
        m_namesSeparator = m_originalNamesSeparator;
        m_originalNamesSeparator.clear();
        return true;
    } else {
        return false;
    }
}

const QString &Settings::namesSeparator() const
{
    return m_namesSeparator;
}

// Network
// -------

void Settings::saveServerPort(int port)
{
    setValue(s_network_serverPort, port);
}

int Settings::serverPort() const
{
    const int port = value(s_network_serverPort, DefaultValues::serverPort).toInt();
    return (port >= 1024 && port <= 65535) ? port : DefaultValues::serverPort;
}

void Settings::saveConnectTimeout(int timeout)
{
    setValue(s_network_connectTimeout, timeout);
}

int Settings::connectTimeout() const
{
    return value(s_network_connectTimeout, DefaultValues::connectTimeout).toInt();
}

void Settings::saveRequestTimeout(int timeout)
{
    setValue(s_network_requestTimeout, timeout);
}

int Settings::requestTimeout() const
{
    return value(s_network_requestTimeout, DefaultValues::requestTimeout).toInt();
}

void Settings::saveDiscoverPort(int port)
{
    setValue(s_network_discoverPort, port);
}

int Settings::discoverPort() const
{
    const int port = value(s_network_discoverPort, DefaultValues::discoverServerPort).toInt();
    return (port >= 1024 && port <= 65535) ? port : DefaultValues::discoverServerPort;
}

void Settings::saveDiscoverTimeout(int timeout)
{
    setValue(s_network_discoverTimeout, timeout);
}

int Settings::discoverTimeout() const
{
    return value(s_network_discoverTimeout, DefaultValues::discoverTimeout).toInt();
}

void Settings::saveDiscoverAttempts(int attempts)
{
    setValue(s_network_discoverAttempts, attempts);
}

int Settings::discoverAttempts() const
{
    return value(s_network_discoverAttempts, DefaultValues::discoverAttempts).toInt();
}

void Settings::savePreferIpv6(bool state)
{
    setValue(s_network_preferIpv6, state);
}

bool Settings::preferIpv6() const
{
    return value(s_network_preferIpv6, false).toBool();
}

// Draw settings

void Settings::saveDrawSettingsTemplate(const Draw::Settings &data)
{
    setValue(s_drawSettingsTemplate_version, Draw::settingsVersion);
    setValue(s_drawSettingsTemplate_data,    JsonHelper::drawSettingsToString(data));
}

Draw::Settings Settings::drawSettingsTemplate() const
{
    if (value(s_drawSettingsTemplate_version, 0).toInt() != Draw::settingsVersion) {
        return Draw::defaultSettings;
    }

    return JsonHelper::parseDrawSettings(value(s_drawSettingsTemplate_data).toString());
}
