/* SPDX-FileCopyrightText: 2019-2020 Tobias Leupold <tobias.leupold@gmx.de>

   SPDX-License-Identifier: GPL-3.0-or-later
*/

// Local includes
#include "TournamentSettings.h"
#include "JsonHelper.h"

// Qt includes
#include <QDebug>
#include <QJsonObject>
#include <QJsonArray>

// C++ includes
#include <utility>

// The revision we use
static constexpr int s_tournamentSettingsVersion = 2;

// JSON Keys for all settings

static const QString s_muckturnierVersion = QStringLiteral("muckturnier version");
static const QString s_tournamentMode     = QStringLiteral("tournament mode");
static const QString s_boogerScore        = QStringLiteral("booger score");
static const QString s_boogersPerRound    = QStringLiteral("boogers per round");

static const QString s_tournamentOpen = QStringLiteral("tournament open");

static const QString s_scoreType               = QStringLiteral("score type");
static const QString s_noTableNumbers          = QStringLiteral("no table numbers");
static const QString s_autoselectPairs         = QStringLiteral("autoselect pairs");
static const QString s_selectByLastRound       = QStringLiteral("select by last round");
static const QString s_includeOpponentGoals    = QStringLiteral("include opponent goals");
static const QString s_grayOutUnimportantGoals = QStringLiteral("gray out unimportant goals");

static const QString s_defaultMarker            = QStringLiteral("default marker");
static const QString s_drawnMarker              = QStringLiteral("drawn marker");
static const QString s_absentMarker             = QStringLiteral("absent marker");
static const QString s_singlesManagementVersion = QStringLiteral("singles management version");
static const QString s_singlesManagement        = QStringLiteral("singles management");

static const QString s_registrationColumns          = QStringLiteral("registration columns");
static const QString s_registrationListHeaderHidden = QStringLiteral("registration list header "
                                                                     "hidden");

static const QString s_drawSettingsVersion = QStringLiteral("draw settings version");
static const QString s_drawSettings        = QStringLiteral("draw settings");

static const QString s_optionalPages = QStringLiteral("optional pages");

TournamentSettings::TournamentSettings(QObject *parent) : QObject(parent)
{
}

int TournamentSettings::version() const
{
    return s_tournamentSettingsVersion;
}

bool TournamentSettings::loadingDb() const
{
    return m_loadingDb;
}

void TournamentSettings::setLoadingDb(bool state)
{
    m_loadingDb = state;
}

QString TournamentSettings::updateSettingsFromV8(const QString &data) const
{
    // Move the "players page columns" hash to the "registration columns" array

    const QString playersPageColumns = QStringLiteral("players page columns");
    QJsonObject settings = JsonHelper::jsonObjectFromString(data);
    const auto columns = JsonHelper::parsePlayersPageColumns(
                             settings.value(playersPageColumns).toObject());

    settings.insert(QStringLiteral("registration columns"),
                    JsonHelper::registrationColumnsToJson(columns));
    settings.remove(playersPageColumns);

    return JsonHelper::serializeJson(settings);
}

const QVector<OptionalPages::Page> &TournamentSettings::optionalPages() const
{
    return m_optionalPages;
}

void TournamentSettings::setOptionalPageShown(OptionalPages::Page page, bool state)
{
    if (m_loadingDb) {
        return;
    }

    if (state) {
        m_optionalPages.append(page);
    } else {
        m_optionalPages.removeOne(page);
    }
}

void TournamentSettings::setDbSettingsVersion(int version)
{
    m_dbSettingsVersion = version;
}

int TournamentSettings::dbSettingsVersion() const
{
    return m_dbSettingsVersion;
}

bool TournamentSettings::load(const QString &data)
{
    const QJsonObject json = JsonHelper::jsonObjectFromString(data);

    // There are some essential values that have to be present and valid. We parse them here:

    const QString tournamentMode = json.value(s_tournamentMode).toString();
    bool tournamentModeValid = false;
    QHashIterator<Tournament::Mode, QString> it(Tournament::modeMap);
    while (it.hasNext()) {
        it.next();
        if (it.value() == tournamentMode) {
            m_tournamentMode = it.key();
            tournamentModeValid = true;
            break;
        }
    }
    if (! tournamentModeValid) {
        return false;
    }

    QJsonValue value;

    value = json.value(s_boogerScore);
    if (value.type() == QJsonValue::Double) {
        m_boogerScore = value.toInt();
    } else {
        return false;
    }

    value = json.value(s_boogersPerRound);
    if (value.type() == QJsonValue::Double) {
        m_boogersPerRound = value.toInt();
    } else {
        return false;
    }

    // For all other values, we can fall back to a default setting

    const QJsonValue tournamentOpenValue(json.value(s_tournamentOpen));
    m_tournamentOpen = tournamentOpenValue.type() == QJsonValue::Bool
                           ? tournamentOpenValue.toBool() : true;
    m_originalTournamentOpen = m_tournamentOpen;

    const QString muckturnierVersion = json.value(s_muckturnierVersion).toString();
    m_muckturnierVersion = muckturnierVersion.isEmpty() ? tr("unbekannt") : muckturnierVersion;

    m_noTableNumbers = json.value(s_noTableNumbers).toBool();

    m_autoSelectPairs = json.value(s_autoselectPairs).toBool();
    m_selectByLastRound = json.value(s_selectByLastRound).toBool();

    m_includeOpponentGoals = json.value(s_includeOpponentGoals).toBool();

    m_grayOutUnimportantGoals = json.value(s_grayOutUnimportantGoals).toBool();

    m_marker[Markers::Default] = json.value(s_defaultMarker).toInt();

    const auto drawnMarkerValue = json.value(s_drawnMarker);
    m_marker[Markers::Drawn] = drawnMarkerValue.isUndefined() ? -1 : drawnMarkerValue.toInt();

    m_marker[Markers::Absent] = json.value(s_absentMarker).toInt();

    const bool singlesManagementParsed = json.value(s_singlesManagementVersion).toInt()
                                         == Markers::singlesManagementVersion;
    m_singlesManagement = singlesManagementParsed
        ? JsonHelper::parseSinglesManagement(json.value(s_singlesManagement).toObject())
        : Markers::defaultSinglesManagement;

    m_registrationColumns = JsonHelper::parseRegistrationColumns(
                                json.value(s_registrationColumns).toArray());

    m_registrationListHeaderHidden = json.value(s_registrationListHeaderHidden).toBool();

    m_scoreType = Tournament::scoreTypeMap.key(json.value(s_scoreType).toString(),
                                                 Tournament::ScoreType::Horizontal);

    const bool drawSettingsParsed = json.value(s_drawSettingsVersion).toInt()
                                        == Draw::settingsVersion;
    m_drawSettings = drawSettingsParsed
        ? JsonHelper::parseDrawSettings(json.value(s_drawSettings).toObject())
        : Draw::defaultSettings;

    m_optionalPages = JsonHelper::parseOptionalPages(json.value(s_optionalPages).toArray());

    m_unhandledValues.clear();
    if (m_dbSettingsVersion > s_tournamentSettingsVersion) {
        // The settings were created by a newer version of Muckturnier. Some keys are not handled by
        // this version yet. To keep them when closing and saving the currrent settings, we cache
        // them here so that they can be re-added later.

        auto unhandledKeys = json.keys();
        unhandledKeys.removeOne(s_muckturnierVersion);
        unhandledKeys.removeOne(s_tournamentMode);
        unhandledKeys.removeOne(s_tournamentOpen);
        unhandledKeys.removeOne(s_scoreType);
        unhandledKeys.removeOne(s_boogerScore);
        unhandledKeys.removeOne(s_boogersPerRound);
        unhandledKeys.removeOne(s_noTableNumbers);
        unhandledKeys.removeOne(s_autoselectPairs);
        unhandledKeys.removeOne(s_selectByLastRound);
        unhandledKeys.removeOne(s_includeOpponentGoals);
        unhandledKeys.removeOne(s_grayOutUnimportantGoals);
        unhandledKeys.removeOne(s_defaultMarker);
        unhandledKeys.removeOne(s_drawnMarker);
        unhandledKeys.removeOne(s_absentMarker);
        if (singlesManagementParsed) {
            unhandledKeys.removeOne(s_singlesManagementVersion);
            unhandledKeys.removeOne(s_singlesManagement);
        }
        unhandledKeys.removeOne(s_registrationColumns);
        unhandledKeys.removeOne(s_registrationListHeaderHidden);
        if (drawSettingsParsed) {
            unhandledKeys.removeOne(s_drawSettingsVersion);
            unhandledKeys.removeOne(s_drawSettings);
        }
        unhandledKeys.removeOne(s_optionalPages);

        for (const QString &key : std::as_const(unhandledKeys)) {
            m_unhandledValues.insert(key, json.value(key));
        }
    }

    return true;
}

QString TournamentSettings::toString() const
{
    QJsonObject settings {
        { s_muckturnierVersion,           m_muckturnierVersion },
        { s_tournamentMode,               Tournament::modeMap.value(m_tournamentMode) },
        { s_tournamentOpen,               m_tournamentOpen },
        { s_scoreType,                    Tournament::scoreTypeMap.value(m_scoreType) },
        { s_boogerScore,                  m_boogerScore },
        { s_boogersPerRound,              m_boogersPerRound },
        { s_noTableNumbers,               m_noTableNumbers },
        { s_autoselectPairs,              m_autoSelectPairs },
        { s_selectByLastRound,            m_selectByLastRound },
        { s_includeOpponentGoals,         m_includeOpponentGoals },
        { s_grayOutUnimportantGoals,      m_grayOutUnimportantGoals },
        { s_defaultMarker,                m_marker.value(Markers::Default) },
        { s_drawnMarker,                  m_marker.value(Markers::Drawn) },
        { s_absentMarker,                 m_marker.value(Markers::Absent) },
        { s_singlesManagementVersion,     Markers::singlesManagementVersion },
        { s_singlesManagement,            JsonHelper::singlesManagementToJson(
                                              m_singlesManagement) },
        { s_registrationColumns,          JsonHelper::registrationColumnsToJson(
                                              m_registrationColumns) },
        { s_registrationListHeaderHidden, m_registrationListHeaderHidden },
        { s_drawSettingsVersion,          Draw::settingsVersion },
        { s_drawSettings,                 JsonHelper::drawSettingsToJson(m_drawSettings) },
        { s_optionalPages,                JsonHelper::optionalPagesToJson(m_optionalPages) }
    };

    // If we have unhandled values, add them as well so that they aren't lost
    if (! m_unhandledValues.isEmpty()) {
        QHashIterator<QString, QJsonValue> it(m_unhandledValues);
        while (it.hasNext()) {
            it.next();
            settings.insert(it.key(), it.value());
        }
    }

    return JsonHelper::serializeJson(settings);
}

void TournamentSettings::setMuckturnierVersion(const QString &text)
{
    m_muckturnierVersion = text;
}

const QString &TournamentSettings::muckturnierVersion() const
{
    return m_muckturnierVersion;
}

void TournamentSettings::setTournamentMode(Tournament::Mode tournamentMode)
{
    m_tournamentMode = tournamentMode;
}

Tournament::Mode TournamentSettings::tournamentMode() const
{
    return m_tournamentMode;
}

bool TournamentSettings::isPairMode() const
{
    return m_tournamentMode == Tournament::Mode::FixedPairs;
}

bool TournamentSettings::isSinglePlayerMode() const
{
    return m_tournamentMode == Tournament::Mode::SinglePlayers;
}

void TournamentSettings::setTournamentOpen(bool state)
{
    m_tournamentOpen = state;
}

bool TournamentSettings::tournamentOpen() const
{
    return m_tournamentOpen;
}

bool TournamentSettings::tournamentOpenChanged() const
{
    return m_tournamentOpen != m_originalTournamentOpen;
}

void TournamentSettings::setBoogerScore(int score)
{
    m_boogerScore = score;
}

const int &TournamentSettings::boogerScore() const
{
    return m_boogerScore;
}

void TournamentSettings::setBoogersPerRound(int count)
{
    m_boogersPerRound = count;
}

int TournamentSettings::boogersPerRound() const
{
    return m_boogersPerRound;
}

void TournamentSettings::setNoTableNumbers(bool state)
{
    m_noTableNumbers = state;
}

bool TournamentSettings::noTableNumbers() const
{
    return m_noTableNumbers;
}

void TournamentSettings::setAutoSelectPairs(bool state)
{
    m_autoSelectPairs = state;
}

bool TournamentSettings::autoSelectPairs() const
{
    return m_autoSelectPairs;
}

void TournamentSettings::setSelectByLastRound(bool state)
{
    m_selectByLastRound = state;
}

bool TournamentSettings::selectByLastRound() const
{
    return m_selectByLastRound;
}

void TournamentSettings::setIncludeOpponentGoals(bool state)
{
    m_includeOpponentGoals = state;
}

bool TournamentSettings::includeOpponentGoals() const
{
    return m_includeOpponentGoals;
}

void TournamentSettings::setGrayOutUnimportantGoals(bool state)
{
    m_grayOutUnimportantGoals = state;
}

bool TournamentSettings::grayOutUnimportantGoals() const
{
    return m_grayOutUnimportantGoals;
}

void TournamentSettings::setMarker(Markers::Type type, int id)
{
    m_marker[type] = id;
}

int TournamentSettings::marker(Markers::Type type) const
{
    return m_marker.value(type);
}

void TournamentSettings::setSinglesManagement(const Markers::SinglesManagement &data)
{
    m_singlesManagement = data;
}

const Markers::SinglesManagement &TournamentSettings::singlesManagement() const
{
    return m_singlesManagement;
}

void TournamentSettings::setRegistrationColumns(const QVector<RegistrationColumns::Column> &data)
{
    m_registrationColumns = data;
}

const QVector<RegistrationColumns::Column> &TournamentSettings::registrationColumns() const
{
    return m_registrationColumns;
}

void TournamentSettings::setRegistrationListHeaderHidden(bool state)
{
    m_registrationListHeaderHidden = state;
}

bool TournamentSettings::registrationListHeaderHidden() const
{
    return m_registrationListHeaderHidden;
}

void TournamentSettings::setScoreType(Tournament::ScoreType type)
{
    m_scoreType = type;
}

Tournament::ScoreType TournamentSettings::scoreType() const
{
    return m_scoreType;
}

void TournamentSettings::setDrawSettings(const Draw::Settings &settings)
{
    m_drawSettings = settings;
}

const Draw::Settings &TournamentSettings::drawSettings() const
{
    return m_drawSettings;
}
